import pygame
import pymunk
import math
from typing import Tuple
from .consts import PLAYER_WHITE, PLAYER_BLACK
from ..color import Color


class Stone:
    RADIUS = 15
    MASS = 10
    ELASTICITY = 0.85

    def __init__(self, x: float, y: float, player: int, index: int):
        moment = pymunk.moment_for_circle(Stone.MASS, 0, Stone.RADIUS)

        self._body = pymunk.Body(Stone.MASS, moment)
        self._body.position = (x, y)

        self._shape = pymunk.Circle(self._body, Stone.RADIUS)
        self._shape.elasticity = Stone.ELASTICITY

        self._player = player
        self._index = index

        self.active = True

    def draw(self, surface: pygame.Surface):
        if not self.active:
            return

        font = pygame.font.SysFont(None, 24)
        if self._player == PLAYER_BLACK:
            pygame.draw.circle(surface, Color.BLACK, self._body.position, Stone.RADIUS)  # Black
            text_surface = font.render(str(self._index), True, Color.WHITE)
        else:
            pygame.draw.circle(surface, Color.WHITE, self._body.position, Stone.RADIUS)  # White
            pygame.draw.circle(surface, Color.BLACK, self._body.position, Stone.RADIUS, 1)  # Black outline
            text_surface = font.render(str(self._index), True, Color.BLACK)

        text_rect = text_surface.get_rect(center=self.position_)
        surface.blit(text_surface, text_rect)

    def apply_impulse(self, impulse: Tuple[float, float] = (0, 0)):
        self._body.apply_impulse_at_local_point(impulse, (0, 0))

    def stop(self):
        self._body.velocity = (0, 0)

    @property
    def position_(self):
        return self._body.position

    @property
    def index_(self):
        return self._index

    @property
    def player_(self):
        return self._player

    @property
    def shape_(self):
        return self._shape

    @property
    def body_(self):
        return self._body

    @property
    def velocity_(self):
        return self._body.velocity


class Obstacle:
    ELASTICITY = 0.25

    def __init__(self, x: float, y: float, width: float, height: float):
        self._width = width
        self._height = height

        self._body = pymunk.Body(body_type=pymunk.Body.STATIC)  # Static body (doesn't move)
        self._body.position = (x, y)

        self._shape = pymunk.Poly.create_box(self._body, (width, height))
        self._shape.elasticity = Obstacle.ELASTICITY  # Less bouncy for wooden board feel

    def draw(self, surface: pygame.Surface):
        x, y = self._body.position
        rect = pygame.Rect(x - self._width / 2, y - self._height / 2, self._width, self._height)
        pygame.draw.rect(surface, Color.BRASS, rect)
        pygame.draw.rect(surface, Color.BLACK, rect, 1)  # Black outline

    @property
    def size_(self):
        return self._width, self._height

    @property
    def position_(self):
        return self._body.position

    @property
    def shape_(self):
        return self._shape

    @property
    def body_(self):
        return self._body


class SlingShot:
    POWER_LINE_LENGTH = 100

    def __init__(self,
                 pos_start: Tuple[float, float] | pymunk.Vec2d | None = None,
                 pos_end: Tuple[float, float] | pymunk.Vec2d | None = None):
        self.pos_start = pos_start
        self.pos_end = pos_end
        self.wait = False

    def draw(self, surface: pygame.Surface):
        if self.pos_start is None or self.pos_end is None:
            return

        sx, sy = self.pos_start
        ex, ey = self.pos_end
        dx, dy = sx - ex, sy - ey
        distance = math.hypot(dx, dy)

        if distance <= 0:
            return

        nx, ny = dx / distance, dy / distance

        len_arrow, len_dash = 75, 5
        n_dashes = int(len_arrow / (2 * len_dash))

        ex, ey = sx + nx * len_arrow, sy + ny * len_arrow

        for i in range(n_dashes):
            start_ratio = i * (2 * len_dash) / len_arrow
            end_ratio = min(start_ratio + len_dash / len_arrow, 1)
            dash_start = (
                sx + (ex - sx) * start_ratio,
                sy + (ey - sy) * start_ratio
            )
            dash_end = (
                sx + (ex - sx) * end_ratio,
                sy + (ey - sy) * end_ratio
            )
            pygame.draw.line(surface, Color.LIGHT_GREY, dash_start, dash_end, 2)

        arrow_size = 10
        arrow_angle = 30  # degrees

        # Calculate arrow head points
        angle = pygame.math.Vector2(0, 0).angle_to(pygame.math.Vector2(nx, ny))
        l_angle, r_angle = math.radians(angle + arrow_angle), math.radians(angle - arrow_angle)

        arrow_left = (
            ex - arrow_size * math.cos(l_angle),
            ey - arrow_size * math.sin(l_angle)
        )
        arrow_right = (
            ex - arrow_size * math.cos(r_angle),
            ey - arrow_size * math.sin(r_angle)
        )

        pygame.draw.line(surface, Color.LIGHT_GREY, (ex, ey), arrow_left, 2)
        pygame.draw.line(surface, Color.LIGHT_GREY, (ex, ey), arrow_right, 2)

        len_power = min(distance, SlingShot.POWER_LINE_LENGTH)
        power_end = (
            sx - nx * len_power,
            sy - ny * len_power
        )

        pygame.draw.line(surface, Color.LIGHT_GREY, (sx, sy), power_end, 3)



import gymnasium as gym
import numpy as np
import pygame
from .env import BulletBillEnv
from .consts import *
from ..manual import ManualPlayWrapper


class RGBImgObsWrapper(gym.ObservationWrapper):
    def __init__(self, env):
        super().__init__(env)
        env = env.unwrapped
        assert isinstance(env, BulletBillEnv)
        self.observation_space = gym.spaces.Box(low=0, high=255, shape=(env.height, env.width, 3), dtype=np.uint8)

    def observation(self, obs):
        env = self.env.unwrapped
        assert isinstance(env, BulletBillEnv)
        return env.get_frame()


class BulletBillManualPlayWrapper(ManualPlayWrapper):
    KEY_TO_CONTINUOUS_ACTION = {
        pygame.K_LEFT: np.array([-MARIO_DEFAULT_MOVING_FORCE, 0]),
        pygame.K_RIGHT: np.array([MARIO_DEFAULT_MOVING_FORCE, 0]),
        pygame.K_SPACE: np.array([0, MARIO_JUMPING_FORCE]),
    }

    KEY_TO_DISCRETE_ACTION = {
        pygame.K_LEFT: Actions.left,
        pygame.K_RIGHT: Actions.right,
        pygame.K_SPACE: Actions.jump
    }

    def __init__(self, env, *args, **kwargs):
        super().__init__(env, **kwargs)
        env = self.env.unwrapped
        assert isinstance(env, BulletBillEnv)
        self._continuous_action = env.continuous_action
        self._keys_to_action = self.KEY_TO_CONTINUOUS_ACTION if self._continuous_action else self.KEY_TO_DISCRETE_ACTION

    def handle_events(self, event):
        if event.type == pygame.KEYDOWN:
            return self._keys_to_action.get(event.key)
        return None

    @property
    def default_action_(self):
        if self._continuous_action:
            return np.array([0.0, 0.0])
        else:
            return Actions.noop



from typing import Tuple, List, Optional
import gymnasium as gym
import numpy as np
import pygame
import pymunk
import random
import math
from .consts import *
from .objs import Stone, Obstacle, SlingShot
from ..color import Color


class AlkkagiEnv(gym.Env):
    metadata = {
        'render_modes': ['human', 'rgb_array'],
        'render_fps': 60,
    }

    def __init__(self, n_stones: int, n_obstacles: int, render_mode: str):
        self.n_stones = n_stones
        self.n_obstacles = n_obstacles
        self.render_mode = render_mode

        self._space = pymunk.Space()
        self._space.damping = DAMPING

        self._stones: List[Stone] = []
        self._obstacles: List[Obstacle] = []

        self._turn = PLAYER_BLACK
        self._is_stone_in_motion = False

        self._screen = None
        self._game_surface = None
        self._status_surface = None
        self._clock = None

        self._slingshot = SlingShot()
        self._steps = 0

        self.width = GAME_WIDTH
        self.height = GAME_HEIGHT
        self.max_power = MAX_POWER

        self.observation_space = gym.spaces.Dict({
            'turn': gym.spaces.Discrete(2),
            'black': gym.spaces.Box(
                low=0,
                high=max(self.width, self.height),
                shape=(n_stones, 3),
                dtype=np.float32
            ),
            'white': gym.spaces.Box(
                low=0,
                high=max(self.width, self.height),
                shape=(n_stones, 3),
                dtype=np.float32
            ),
            'obstacles': gym.spaces.Box(
                low=0,
                high=max(self.width, self.height),
                shape=(n_stones, 4),
                dtype=np.float32
            )
        })

        self.action_space = gym.spaces.Dict({
            'turn': gym.spaces.Discrete(2),
            'index': gym.spaces.Box(0, n_stones, shape=(1,), dtype=np.uint8),
            'power': gym.spaces.Box(1, self.max_power, shape=(1,)),
            'angle': gym.spaces.Box(-180, 180, shape=(1,)),
        })

    @property
    def turn_(self) -> int:
        return self._turn

    @property
    def stone_in_motion_(self) -> bool:
        return self._is_stone_in_motion

    @property
    def latest_obs_(self):
        return self._generate_obs()

    @property
    def latest_info_(self):
        return self._generate_info()

    def set_slingshot_from_pos(
            self,
            pos_start: Tuple[float, float] | pymunk.Vec2d | None,
            pos_end: Tuple[float, float] | pymunk.Vec2d | None
    ):
        self._slingshot.pos_start = pos_start
        self._slingshot.pos_end = pos_end

    def set_slingshot_from_angle_power(
            self,
            pos_start: Tuple[float, float] | pymunk.Vec2d | None,
            angle: float,
            power: float
    ):
        sx, sy = pos_start
        dx, dy = math.cos(math.radians(angle)), math.sin(math.radians(angle))
        length = power / self.max_power * SlingShot.POWER_LINE_LENGTH
        ex, ey = sx - dx * length, sy - dy * length
        self.set_slingshot_from_pos((sx, sy), (ex, ey))

    def find_stone(self, pos: Tuple[float, float]) -> Stone:
        mx, my = pos

        closest_stone = None
        min_distance = float('inf')

        for stone in self._stones:
            if stone.player_ != self._turn:
                continue

            x, y = stone.position_
            dx, dy = x - mx, y - my
            distance = math.hypot(dx, dy)

            if distance <= Stone.RADIUS and distance < min_distance:
                closest_stone = stone
                min_distance = distance

        return closest_stone

    def reset(self, **kwargs):
        for stone in self._stones:
            if stone.active:
                self._space.remove(stone.body_, stone.shape_)

        for obstacle in self._obstacles:
            self._space.remove(obstacle.body_, obstacle.shape_)

        self._stones.clear()
        self._obstacles.clear()

        spacing = (self.height - 200) / (self.n_stones - 1)
        x, y = GRID_SIZE * 3, 100

        for i in range(self.n_stones):
            stone = Stone(x, y, PLAYER_BLACK, i)
            self._space.add(stone.body_, stone.shape_)
            self._stones.append(stone)
            y += spacing

        x, y = self.width - GRID_SIZE * 3, 100
        for i in range(self.n_stones):
            stone = Stone(x, y, PLAYER_WHITE, i)
            self._space.add(stone.body_, stone.shape_)
            self._stones.append(stone)
            y += spacing

        if self.n_obstacles > 0:
            spacing = (self.height - 200) / (self.n_obstacles - 1)
            x, y = self.width / 2, 100
            width, height = 10, 80
            for i in range(self.n_obstacles):
                obstacle = Obstacle(x, y, width, height)
                self._space.add(obstacle.body_, obstacle.shape_)
                self._obstacles.append(obstacle)
                y += spacing

        self._turn = PLAYER_BLACK
        self._steps = 0
        self.set_slingshot_from_pos(None, None)
        self.render()
        return self._generate_obs(), self._generate_info()

    def step(self, action):
        turn, index, power, angle = action['turn'], action['index'], action['power'], action['angle']

        if turn != self._turn:
            return self._generate_obs(), 0, False, False, self._generate_info()

        angle = np.clip(angle, -180, 180)
        power = np.clip(power, 1, self.max_power)
        stones = [stone for stone in self._stones if stone.player_ == self._turn and stone.active]
        selected_stone = stones[0]

        for stone in stones:
            if stone.index_ == index:
                selected_stone = stone
                break

        self.set_slingshot_from_angle_power(selected_stone.position_, angle, power)

        dx, dy = math.cos(math.radians(angle)), math.sin(math.radians(angle))
        impulse = power * dx, power * dy

        selected_stone.apply_impulse(impulse)
        self._is_stone_in_motion = True

        while self._is_stone_in_motion:
            self._space.step(1 / self.metadata['render_fps'])

            stones_oob = [
                stone
                for stone in self._stones
                if stone.active and
                   (stone.position_.x < 0 or stone.position_.x > self.width
                   or stone.position_.y < 0 or stone.position_.y > self.height)
            ]

            for stone in stones_oob:
                self._space.remove(stone.body_, stone.shape_)
                stone.active = False

            for stone in self._stones:
                if stone.velocity_.length < VELOCITY_THRESHOLD:
                    stone.stop()

            self._is_stone_in_motion = any(
                stone.active and stone.velocity_.length >= VELOCITY_THRESHOLD for stone in self._stones
            )

            if self.render_mode == 'human':
                self.render()

        self._turn = PLAYER_BLACK if self._turn == PLAYER_WHITE else PLAYER_WHITE
        self.set_slingshot_from_pos(None, None)

        terminated = self._check_win_condition() is not None

        self._steps += 1

        self.render()
        return self._generate_obs(), 0, terminated, False, self._generate_info()

    def render(self,):
        screen_width, screen_height = self.width, self.height + STATUS_HEIGHT
        if self._screen is None:
            pygame.init()
            if self.render_mode == 'human':
                pygame.display.init()
                self._screen = pygame.display.set_mode((screen_width, screen_height))
                pygame.display.set_caption(self.spec.id)
            else:
                self._screen = pygame.Surface((screen_width, screen_height))

            self._game_surface = self._screen.subsurface((0, 0, self.width, self.height))
            self._status_surface = self._screen.subsurface((0, self.height, screen_width, STATUS_HEIGHT))

        if self._clock is None:
            self._clock = pygame.time.Clock()

        self._draw_go_board(self._game_surface)

        for obstacle in self._obstacles:
            obstacle.draw(self._game_surface)

        for stone in self._stones:
            stone.draw(self._game_surface)

        self._slingshot.draw(self._game_surface)

        if self._is_stone_in_motion:
            self._draw_status(
                text=f"({self._steps + 1}-th Action) Stones are moving.",
                text_color=Color.RED,
                fill=Color.RED
            )
        else:
            if self._turn == PLAYER_BLACK:
                self._draw_status(
                    text=f"({self._steps + 1}-th Action) Black player's turn",
                    text_color=Color.BLACK,
                    fill=Color.BLACK
                )
            else:
                self._draw_status(
                    text=f"({self._steps + 1}-th Action) White player's turn",
                    text_color=Color.BLACK,
                    fill=Color.WHITE,
                    stroke=Color.BLACK,
                )

        winner = self._check_win_condition()

        if winner is not None:
            winner = 'Black' if winner == PLAYER_BLACK else 'White'
            self._draw_status(
                text=f"Winner is {winner} player!",
                text_color=Color.BLUE,
                fill=Color.BLUE
            )

        if self.render_mode == 'human':
            pygame.event.pump()
            self._clock.tick(self.metadata['render_fps'])
            pygame.display.flip()
            return None
        else:
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self._screen)), axes=(1, 0, 2)
            )

    def get_frame(self):
        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self._game_surface)), axes=(1, 0, 2)
        )

    def close(self):
        if self._screen:
            pygame.display.quit()

        pygame.quit()

    def _generate_info(self):
        return {
            'steps': self._steps,
        }

    def _generate_obs(self):
        black = np.zeros((self.n_stones, 3), dtype=np.float32)
        white = np.zeros((self.n_stones, 3), dtype=np.float32)

        for stone in self._stones:
            if stone.player_ == PLAYER_BLACK:
                black[stone.index_ ] = (*stone.position_, 1 if stone.active else 0)
            elif stone.player_ == PLAYER_WHITE:
                white[stone.index_] = (*stone.position_, 1 if stone.active else 0)

        obstacles = np.zeros((self.n_obstacles, 4), dtype=np.float32)
        for i, obstacle in enumerate(self._obstacles):
            obstacles[i] = (*obstacle.position_, *obstacle.size_)

        return {
            'turn': self._turn,
            'black': black,
            'white': white,
            'obstacles': obstacles
        }

    def _check_win_condition(self) -> Optional[int]:
        black = sum(1 for stone in self._stones if stone.player_ == PLAYER_BLACK and stone.active)
        white = sum(1 for stone in self._stones if stone.player_ == PLAYER_WHITE and stone.active)

        if black == 0:
           return PLAYER_WHITE
        elif white == 0:
           return PLAYER_BLACK
        else:
           return None

    def _draw_go_board(self, surface: pygame.Surface):
        surface.fill(Color.WOOD)

        for y in range(10, self.height, 20):
            start_x = random.randint(0, 20)
            end_x = self.width - random.randint(0, 20)
            pygame.draw.line(surface, Color.GRAIN, (start_x, y), (end_x, y), 1)

        for x in range(0, self.width, GRID_SIZE):
            pygame.draw.line(surface, Color.BLACK, (x, 0), (x, self.height), 1)

        for y in range(0, self.height, GRID_SIZE):
            pygame.draw.line(surface, Color.BLACK, (0, y), (self.width, y), 1)

    def _draw_status(
            self,
            text: str,
            text_color: Tuple[int, int, int],
            fill: Tuple[int, int, int],
            stroke: Tuple[int, int, int] = None
    ):
        self._status_surface.fill(Color.WHITE)

        rect = pygame.Rect(10, 15, 20, 20)
        pygame.draw.circle(self._status_surface, fill, rect.center, 10)

        if stroke:
            pygame.draw.circle(self._status_surface, stroke, rect.center, 10, 1)  # Black outline

        font = pygame.font.SysFont(None, 24)
        text = font.render(text, True, text_color)
        self._status_surface.blit(text, (rect.right + 10, rect.centery - 8))




import os
from enum import IntEnum

FPS = 15
SCALE = 2.0
TILE_SIZE = 16
GAME_TILE_SIZE = int(TILE_SIZE * SCALE)
STATUS_HEIGHT = 50
SPRITE_COLOR_KEY = (116, 116, 116)
TILESET_WIDTH = 10


class Colors(IntEnum):
    blue = 1
    red = 2
    green = 3
    emerald = 4
    purple = 5
    yellow = 6


class Objects(IntEnum):
    none = -1
    wall = 0
    water = 1
    fire = 2
    link = 3
    stair = 4
    cloud = 5
    sword = 6
    darknut = 7
    goriya = 8
    wizzrobe = 9
    rope = 10
    keese = 11
    moblin = 12
    armos = 13
    tektite = 14
    octorok = 15


class Actions(IntEnum):
    stay = 0
    turn_left = 1
    turn_right = 2
    move_forward = 3
    pick_up = 4
    drop = 5
    attack = 6


class Directions(IntEnum):
    left = 0
    up = 1
    right = 2
    down = 3


class LinkStatus(IntEnum):
    dead = -1
    normal = 0
    picking_up = 1
    attacking = 2
    cleared = 3


class EnemyStatus(IntEnum):
    dead = -1
    normal = 0
    attacking = 1


class CloudStatus(IntEnum):
    disappeared = 0
    appearing_slightly = 1
    appearing_more = 2
    appearing_fully = 3
    disappearing_slightly = 4
    disappearing_more = 5
    disappeared_again = 6


TILESET_ROW_TO_COLOR = {
    2: Colors.blue,
    3: Colors.red,
    4: Colors.green,
    5: Colors.emerald,
    6: Colors.purple,
    7: Colors.yellow
}

TILESET_BASE_TILES = {
    -1: Objects.none,
    0: Objects.wall,
    1: Objects.water,
    2: Objects.water,
    3: Objects.water,
    4: Objects.water,
    5: Objects.water,
    6: Objects.water,
    7: Objects.water,
    8: Objects.water,
    9: Objects.water,
    10: Objects.link,
    11: Objects.stair,
    12: Objects.cloud,
    13: Objects.fire
}

TILESET_OBJECT_START_ID = 20

TILESET_OBJECTS = {
    0: Objects.sword,
    1: Objects.darknut,
    2: Objects.goriya,
    3: Objects.wizzrobe,
    4: Objects.rope,
    5: Objects.keese,
    6: Objects.moblin,
    7: Objects.armos,
    8: Objects.tektite,
    9: Objects.octorok,
}

PALETTE_ENEMY = {
    Colors.blue: [
        (92, 148, 252),
        (0, 0, 168)
    ],
    Colors.red: [
        (252, 152, 56),
        (168, 0, 0)
    ],
    Colors.green: [
        (0, 168, 0),
        (0, 80, 0)
    ],
    Colors.emerald: [
        (88, 248, 152),
        (0, 60, 20)
    ],
    Colors.purple: [
        (200, 147, 222),
        (187, 60, 240)
    ],
    Colors.yellow: [
        (240, 188, 60),
        (64, 44, 0)
    ]
}

PALETTE_SWORD = {
    Colors.blue: (92, 148, 152),
    Colors.red: (216, 40, 0),
    Colors.green: (0, 80, 0),
    Colors.emerald: (0, 232, 216),
    Colors.purple: (187, 60, 240),
    Colors.yellow: (170, 117, 0)
}

PATH_SPRITE = os.path.join(os.path.dirname(__file__), 'assets', 'sprite.png')
PATH_TILESET = os.path.join(os.path.dirname(__file__), 'assets', 'tilesets.png')

LINK_ATTACK_CRITICAL_DAMAGE = 1.0
LINK_ATTACK_NORMAL_DAMAGE = 0.25
LINK_ATTACK_BARE_DAMAGE = 0.2

DEFAULT_ANIM_INTERVAL = 0.3
ATTACK_ANIM_INTERVAL = 0.05

LINK_OFFSET_DOWN_IDLE = [(0, 0), (16, 0)]
LINK_OFFSET_RIGHT_IDLE = [(32, 0), (48, 0)]
LINK_OFFSET_UP_IDLE = [(64, 0), (80, 0)]

LINK_OFFSET_DOWN_ACTION = [(96, 0)]
LINK_OFFSET_RIGHT_ACTION = [(112, 0)]
LINK_OFFSET_UP_ACTION = [(128, 0)]

LINK_OFFSET_PICK_UP = [(144, 0), (160, 0)]

LINK_OFFSET_DOWN_ATTACK = [
    (0, 16), (16, 16, 16, 27), (32, 16, 16, 23), (48, 16, 16, 19)
]

LINK_OFFSET_RIGHT_ATTACK = [
    (64, 16), (80, 16, 27, 16), (107, 16, 23, 16), (130, 16, 19, 16)
]

LINK_OFFSET_UP_ATTACK = [
    (149, 28), (165, 16, 16, 28), (181, 17, 16, 27), (197, 25, 16, 19)
]

DARKNUT_OFFSET_DOWN = [(0, 48), (16, 48)]
DARKNUT_OFFSET_UP = [(32, 48), (48, 48)]
DARKNUT_OFFSET_RIGHT = [(64, 48), (80, 48)]
DARKNUT_HIT_POINT = 3

GORIYA_OFFSET_DOWN = [(96, 48), (112, 48)]
GORIYA_OFFSET_UP = [(128, 48), (144, 48)]
GORIYA_OFFSET_RIGHT = [(160, 48), (176, 48)]
GORIYA_HIT_POINT = 2

WIZZROBE_OFFSET_DOWN = [(0, 64), (16, 64)]
WIZZROBE_OFFSET_UP = [(32, 64), (48, 64)]
WIZZROBE_HIT_POINT = 1

ROPE_OFFSET_RIGHT = [(64, 64), (80, 64)]
ROPE_HIT_POINT = 1

MOBLIN_OFFSET_DOWN = [(96, 64), (112, 64)]
MOBLIN_OFFSET_UP = [(128, 64), (144, 64)]
MOBLIN_OFFSET_RIGHT = [(160, 64), (176, 64)]
MOBLIN_HIT_POINT = 1

ARMOS_OFFSET_DOWN = [(0, 80), (16, 80)]
ARMOS_OFFSET_UP = [(32, 80), (48, 80)]
ARMOS_HIT_POINT = 5

OCTOROK_OFFSET = [(64, 80), (80, 80)]
OCTOROK_HIT_POINT = 1

KEESE_OFFSET = [(96, 80), (112, 80)]
KEESE_HIT_POINT = 1

TEKTITE_OFFSET = [(128, 80), (144, 80)]
TEKTITE_HIT_POINT = 1

CLOUD_OFFSET = [(192, 80), (176, 80), (160, 80)]
CLOUD_UNIT_DURATION = 2

DIRECTION_TO_VECTOR = {
    Directions.left: (-1, 0),
    Directions.right: (1, 0),
    Directions.up: (0, -1),
    Directions.down: (0, 1)
}




from enum import IntEnum
import gymnasium as gym
import numpy as np
import pygame
import pygame.freetype
from itertools import product
from minigrid.core.grid import Grid
from minigrid.core.mission import MissionSpace
from minigrid.core.world_object import Lava, Key, Door, Goal, Wall, Ball
from minigrid.minigrid_env import MiniGridEnv


MISSION_NAME = 'Grid Adventure!'

SYM_AGENT = 'A'
SYM_START = 'S'
SYM_GOAL = 'G'
SYM_WALL = 'W'
SYM_LAVA = 'L'
SYM_KEY = 'K'
SYM_DOOR = 'D'
SYM_BALL = 'B'
SYM_EMTPY = 'E'

GENERATABLE_OBJECTS = [
    SYM_WALL,
    SYM_LAVA,
    SYM_DOOR,
    SYM_KEY,
    SYM_BALL
]

STATE_LOCKED = 'L'
STATE_OPEN = 'O'
STATE_CLOSED = 'C'

COLOR_RED = 'R'
COLOR_BLUE = 'B'
COLOR_GREEN = 'G'
COLOR_YELLOW = 'Y'
COLOR_PURPLE = 'P'

DIR_RIGHT = 'R'
DIR_DOWN = 'D'
DIR_LEFT = 'L'
DIR_UP = 'U'

COLOR_TO_ARGUMENT = {
    COLOR_RED: 'red',
    COLOR_BLUE: 'blue',
    COLOR_GREEN: 'green',
    COLOR_YELLOW: 'yellow',
    COLOR_PURPLE: 'purple',
}


class Actions(IntEnum):
    left = 0
    right = 1
    forward = 2
    pickup = 3
    drop = 4
    unlock = 5


class GridWorldEnv(MiniGridEnv):
    """
    Actions available:
    - 0: turn left
    - 1: turn right
    - 2: move forward
    - 3: pick up the key
    - 4: drop the key
    - 5: unlock the door when the key is present
    """
    metadata = {
        'render_modes': ['human', 'rgb_array'],
        'render_fps': 10,
    }

    def __init__(self, max_steps: int, blueprint: np.ndarray, render_mode: str, **kwargs):
        width, height = blueprint.shape
        start_pos = np.argwhere(blueprint == SYM_START)[0]
        assert len(start_pos) == 2, 'Only one start position should be provided.'

        goal_pos = np.argwhere(blueprint == SYM_GOAL)[0]
        assert len(goal_pos) == 2, 'Only one goal position should be provided.'

        super().__init__(
            mission_space=MissionSpace(lambda : MISSION_NAME),
            width=width,
            height=height,
            max_steps=max_steps,
            **kwargs
        )
        self.render_mode = render_mode

        self.obstacles = []

        self.blueprint = blueprint
        self.goal_pos = goal_pos
        self.agent_start_pos = start_pos
        self.agent_start_dir = 0
        self.action_space = gym.spaces.Discrete(len(Actions))

    @property
    def steps_(self):
        return self.step_count

    @property
    def distance_(self):
        x_a, y_a = self.agent_pos
        x_g, y_g = self.goal_pos
        return abs(x_a - x_g) + abs(y_a - y_g)

    def _gen_grid(self, width: int, height: int) -> None:
        self.grid = Grid(width, height)

        for x, y in product(range(width), range(height)):
            symbol= str(self.blueprint[x, y])
            symbol_type = symbol[0] if len(symbol) > 1 else symbol

            if symbol_type not in GENERATABLE_OBJECTS:
                continue

            if symbol_type == SYM_WALL:
                self.grid.set(x, y, Wall())
            elif symbol_type == SYM_LAVA:
                self.put_obj(Lava(), x, y)
            elif symbol_type == SYM_BALL:
                obs = Ball()
                self.put_obj(obs, x, y)
                self.obstacles.append(obs)
            elif symbol_type == SYM_KEY:
                color = COLOR_TO_ARGUMENT[symbol[1]] or 'red'
                self.put_obj(Key(color=color), x, y)
            elif symbol_type == SYM_DOOR:
                color = COLOR_TO_ARGUMENT[symbol[1]] or 'red'
                is_open = False
                is_locked = False

                if symbol[2] == STATE_OPEN:
                    is_open = True
                elif symbol[2] == STATE_LOCKED:
                    is_locked = True
                self.put_obj(Door(color=color, is_open=is_open, is_locked=is_locked), x, y)

        self.put_obj(Goal(), int(self.goal_pos[0]), int(self.goal_pos[1]))
        self.agent_pos = self.agent_start_pos
        self.agent_dir = self.agent_start_dir

    def _reward(self) -> float:
        return 0

    def _gen_info(self):
        return {'steps': self.steps_, 'distance': self.distance_}

    def reset(self, **kwargs):
        obs, _ = super().reset(**kwargs)
        self.render()
        return obs, self._gen_info()

    def step(self, action):
        front_obj = self.grid.get(*self.front_pos)
        is_crashed = front_obj and front_obj.type == "ball"

        for i_obst in range(len(self.obstacles)):
            x_old, y_old = self.obstacles[i_obst].cur_pos
            x_top, y_top = x_old - 1, y_old - 1

            try:
                self.place_obj(
                    self.obstacles[i_obst], top=(x_top, y_top), size=(3, 3), max_tries=100
                )
                self.grid.set(x_old, y_old, None)
            except RecursionError:
                pass

        obs, reward, terminated, truncated, _ = super().step(action)

        if action == Actions.forward and is_crashed:
            terminated = True

        return obs, reward, terminated, truncated, self._gen_info()

    def render(self):
        img = self.get_frame(self.highlight, self.tile_size, False)
        img = np.transpose(img, axes=(1, 0, 2))

        if self.render_size is None:
            self.render_size = img.shape[:2]

        if self.window is None:
            pygame.init()
            if self.render_mode == "human":
                pygame.display.init()
                self.window = pygame.display.set_mode(
                    (self.screen_size, self.screen_size)
                )
                pygame.display.set_caption(self.spec.id)
            else:
                self.window = pygame.Surface((self.screen_size, self.screen_size))

        if self.clock is None:
            self.clock = pygame.time.Clock()

        surf = pygame.surfarray.make_surface(img)
        offset = surf.get_size()[0] * 0.1
        bg = pygame.Surface(
            (int(surf.get_size()[0] + offset), int(surf.get_size()[1] + offset))
        )
        bg.fill((255, 255, 255))
        bg.blit(surf, (offset / 2, 0))

        bg = pygame.transform.smoothscale(bg, (self.screen_size, self.screen_size))

        font_size = 22
        font = pygame.freetype.SysFont(pygame.font.get_default_font(), font_size)
        text = f'Steps: {self.steps_} / Dist.: {self.distance_}'
        text_rect = font.get_rect(text, size=font_size)
        text_rect.center = bg.get_rect().center
        text_rect.y = int(bg.get_height() - font_size * 1.5)
        font.render_to(bg, text_rect, text, size=font_size)

        self.window.blit(bg, (0, 0))

        if self.render_mode == "human":
            pygame.event.pump()
            self.clock.tick(self.metadata["render_fps"])
            pygame.display.flip()
            return None
        elif self.render_mode == "rgb_array":
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self.window)), axes=(1, 0, 2)
            )
        return None



PLAYER_BLACK = 0
PLAYER_WHITE = 1

GAME_WIDTH = 600
GAME_HEIGHT = 600
STATUS_HEIGHT = 50

MAX_POWER = 2500
GRID_SIZE = 600 // 20
DAMPING = 0.6  # Velocity damping factor (lower value = more damping)
VELOCITY_THRESHOLD = 5.0  # Threshold below which stones are considered "at rest"


from abc import ABC, abstractmethod
from typing import Any, Dict


class Agent(ABC):
    @abstractmethod
    def act(self, observation: Any, info: Dict):
        raise NotImplementedError()

    @abstractmethod
    def save(self, path: str):
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def load(cls, path: str) -> 'Agent':
        raise NotImplementedError()


import gymnasium as gym
import numpy as np
import pygame
from .env import ZeldaAdventureEnv
from .consts import *
from ..manual import ManualPlayWrapper


class RGBImgObsWrapper(gym.ObservationWrapper):
    def __init__(self, env):
        super().__init__(env)
        env = env.unwrapped
        assert isinstance(env, ZeldaAdventureEnv)
        self.observation_space = gym.spaces.Box(low=0, high=255, shape=(env.height, env.width, 3), dtype=np.uint8)

    def observation(self, obs):
        env = self.env.unwrapped
        assert isinstance(env, ZeldaAdventureEnv)
        return env.get_frame()


class ZeldaAdventureEnvManualPlayWrapper(ManualPlayWrapper):
    KEY_TO_DISCRETE_ACTION = {
        pygame.K_LEFT: Actions.turn_left,
        pygame.K_RIGHT: Actions.turn_right,
        pygame.K_UP: Actions.move_forward,
        pygame.K_q: Actions.attack,
        pygame.K_w: Actions.pick_up,
        pygame.K_e: Actions.drop,
        pygame.K_r: Actions.stay,
    }

    def __init__(self, env, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        env = self.env.unwrapped
        assert isinstance(env, ZeldaAdventureEnv)
        self._keys_to_action = self.KEY_TO_DISCRETE_ACTION

    def handle_events(self, event):
        if event.type == pygame.KEYDOWN:
            return self.KEY_TO_DISCRETE_ACTION.get(event.key)
        return None

    @property
    def default_action_(self):
        return None



import os
from enum import IntEnum

FPS = 60
SCALE = 3.0
TILE_SIZE = 16
GAME_TILE_SIZE = int(TILE_SIZE * SCALE)
STATUS_HEIGHT = 50
SPRITE_COLOR_KEY = (146, 144, 255)


class Ids(IntEnum):
    mario = 0
    land = 32


class Actions(IntEnum):
    noop = 0
    left = 1
    right = 2


MARIO_ZERO_SPEED_THRESHOLD = 10.0
MARIO_MAX_SPEED = 750.0
MARIO_MIN_ANIM_INTERVAL, MARIO_MAX_ANIM_INTERVAL = 0.005, 0.1
MARIO_FRICTION = 0.90
MARIO_DEFAULT_MOVING_FORCE = 75.0
MARIO_MAX_MOVING_FORCE = 100.0

BLURP_GRAVITY_MIN, BLURP_GRAVITY_MAX = 100, 800
BLURP_INTERVAL_ANIM = 0.1

PATH_SPRITE = os.path.join(os.path.dirname(__file__), 'assets', 'sprite.png')
FILE_OBJECT = '_object.csv'
FILE_BACKGROUND = '_background.csv'

OFFSET_MARIO_STAND = 0, 0
OFFSET_MARIO_WALK = [(16, 0), (32, 0), (48, 0)]
OFFSET_MARIO_DEAD = 160, 0

OFFSET_BLURP = [(0, 16), (16, 16)]

OFFSET_FLOOR = 96, 32




import os
import gymnasium as gym
from .env import BulletBillEnv
from .wrappers import RGBImgObsWrapper
from ..util import play_bgm, ObsType


_BGM_PATH = os.path.join(os.path.dirname(__file__), 'assets', 'bgm.ogg')


def _create_env(
        game_duration: float = 180,
        init_spawn_interval: float = 1,
        min_spawn_interval: float = 0.2,
        max_spawns: int = 30,
        max_spawn_duration: float = 150,
        bgm: bool = False,
        obs_type: ObsType = 'custom',
        continuous_action: bool = False,
        stage: int = 1,
        **kwargs
) -> gym.Env:
    if bgm:
        play_bgm(_BGM_PATH)

    env = BulletBillEnv(
        game_duration=game_duration,
        init_spawn_interval=init_spawn_interval,
        min_spawn_interval=min_spawn_interval,
        max_spawns=max_spawns,
        max_spawn_duration=max_spawn_duration,
        continuous_action=continuous_action,
        stage=stage,
        **kwargs
    )

    if obs_type == 'image':
        env = RGBImgObsWrapper(env)
    
    return env


import pygame
import numpy as np
from typing import Tuple, List
from .consts import *
from ..util import load_sprite
from ..sprite import Sprite


class Mario(Sprite):
    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()

        self._walk_idx = 0
        self._walk_timer = 0
        self._active = True
        self._is_right = True

        images = dict()
        mapper = {
            'right_stand': (OFFSET_MARIO_STAND, False),
            'left_stand': (OFFSET_MARIO_STAND, True),
            'right_walk': (OFFSET_MARIO_WALK, False),
            'left_walk': (OFFSET_MARIO_WALK, True),
            'dead': (OFFSET_MARIO_DEAD, False)
        }
        for key, (offsets, flip_x) in mapper.items():
            size = TILE_SIZE, TILE_SIZE
            if isinstance(offsets, list):
                images[key] = [
                    load_sprite(sprite, size, offset, flip_x, scale=SCALE)
                    for offset in offsets
                ]
            else:
                images[key] = load_sprite(sprite, size, offsets, flip_x, scale=SCALE)
        self._images = images
        self._velocity = 0.0

        self.image = self._images['right_stand']
        self.rect = self.image.get_rect(x=position[0], y=position[1])

    @property
    def velocity_(self):
        return self._velocity

    @property
    def active_(self):
        return self._active

    def apply_force(self, force: float):
        force = np.clip(force, -MARIO_DEFAULT_MOVING_FORCE, MARIO_DEFAULT_MOVING_FORCE)
        self._velocity += force

    def frame_tick(self, dt: float, screen_size: Tuple[int, int]):
        self._velocity = np.clip(self._velocity, -MARIO_MAX_SPEED, MARIO_MAX_SPEED)

        if self._velocity > 0:
            self._is_right = True
        elif self._velocity < 0:
            self._is_right = False

        if abs(self._velocity) < MARIO_ZERO_SPEED_THRESHOLD:
            self._velocity = 0.0

        self.rect.centerx += self._velocity * dt
        self.rect.left = max(0, self.rect.left)
        self.rect.right = min(screen_size[0], self.rect.right)

        self._velocity *= MARIO_FRICTION

    def update(self, dt: float, *args, **kwargs):
        if not self._active:
            self.image = self._images['dead']
        elif abs(self._velocity) >= MARIO_ZERO_SPEED_THRESHOLD:
            self._walk_timer += dt
            speed_factor = abs(self._velocity) / MARIO_MAX_SPEED
            delay = MARIO_MAX_ANIM_INTERVAL - (MARIO_MAX_ANIM_INTERVAL - MARIO_MIN_ANIM_INTERVAL) * speed_factor

            if self._walk_timer >= delay:
                self._walk_timer = 0
                self._walk_idx = (self._walk_idx + 1) % len(self._images['right_walk'])
            self.image = self._images['right_walk'][self._walk_idx] if self._is_right else self._images['left_walk'][self._walk_idx]
        else:
            self.image = self._images['right_stand'] if self._is_right else self._images['left_stand']

    def check_collision(self, blurps: List['Blurp']):
        for blurp in blurps:
            if blurp.rect.colliderect(self.rect):
                self._active = False
                break


class Blurp(Sprite):
    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[float, float],
            gravity: float,
            flip: bool = False,
    ):
        super().__init__()

        self._gravity = gravity
        self._anim_idx = 0
        self._anim_timer = 0
        self._velocity = pygame.Vector2(0, 0)
        self._active = True

        self._images = [
            load_sprite(sprite, (TILE_SIZE, TILE_SIZE), offset, flip, scale=SCALE)
            for offset in OFFSET_BLURP
        ]

        self.image = self._images[0]
        self.rect = self.image.get_rect(x=position[0], y=position[1])

    @property
    def velocity_(self):
        return self._velocity

    @property
    def gravity_(self):
        return self._gravity

    @property
    def active_(self):
        return self._active

    def frame_tick(self, dt: float, screen_size: Tuple[int, int], *args, **kwargs):
        self._velocity.y += self._gravity * dt
        self.rect.centerx += self._velocity.x * dt
        self.rect.centery += self._velocity.y * dt

        self._active = self.rect.y < screen_size[1]

    def update(self, dt: float, *args, **kwargs):
        self._anim_timer += dt
        if self._anim_timer >= BLURP_INTERVAL_ANIM:
            self._anim_timer = 0
            self._anim_idx = (self._anim_idx + 1) % len(self._images)

        self.image = self._images[self._anim_idx]

    def calc_velocity(self, target: Tuple[int, int] = None):
        if target is not None:
            sx, sy = self.rect.center
            ex, ey = target
            dx = ex - sx
            dist = abs(dx)

            angle = np.random.uniform(45, 85)
            rad = np.radians(angle)
            try:
                v_square = (dist * self._gravity) / np.sin(2 * rad)
                speed = np.sqrt(v_square)
            except ValueError:
                speed = 0
            direction = 1.0 if dx > 0 else -1.0
            self._velocity.x = speed * np.cos(rad) * direction
            self._velocity.y = -speed * np.sin(rad)


class WorldObjects(Sprite):
    def __init__(
            self,
            image: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()

        self.image = image
        self.rect = self.image.get_rect(x=position[0], y=position[1])

    def update(self, *args, **kwargs):
        pass

    def frame_tick(self, *args, **kwargs):
        pass


from typing import Tuple, Literal
import numpy as np
import pygame
import pygame.freetype
import gymnasium as gym
from .objs import Link, Cloud, Sword, Darknut, Goriya, Wizzrobe, Keese, Moblin, Rope, Octorok, Armos, Tektite, Stair, \
    Wall, Fire, Water, Enemy
from .consts import *
from ..color import Color
from ..sprite import load_tile_map_data, Group


class ZeldaAdventureEnv(gym.Env):
    metadata = {
        'render_modes': ['human', 'rgb_array'],
        'render_fps': FPS
    }

    enemy_to_inst = {
        Objects.darknut: Darknut,
        Objects.goriya: Goriya,
        Objects.wizzrobe: Wizzrobe,
        Objects.armos: Armos,
        Objects.octorok: Octorok,
        Objects.tektite: Tektite,
        Objects.keese: Keese,
        Objects.moblin: Moblin,
        Objects.rope: Rope,
    }

    def __init__(
            self,
            max_steps: float,
            stage: int = 1,
            render_mode=None
    ):
        self.render_mode = render_mode
        self.stage = stage
        self.max_steps = max_steps

        self._steps = 0
        self._game_state: Literal['init', 'playing', 'game_over', 'cleared'] = 'init'

        sprite = pygame.image.load(PATH_SPRITE)
        sprite.set_colorkey(SPRITE_COLOR_KEY)
        sprite = pygame.transform.scale(
            sprite,
            (sprite.get_width() * SCALE, sprite.get_height() * SCALE)
        )
        self._sprite = sprite

        tileset = pygame.image.load(PATH_TILESET)
        tileset.set_colorkey(SPRITE_COLOR_KEY)
        tileset = pygame.transform.scale(
            tileset,
            (tileset.get_width() * SCALE, tileset.get_height() * SCALE)
        )
        self._tileset = tileset

        self._screen = None
        self._game_surface = None
        self._status_surface = None
        self._clock = pygame.time.Clock()

        self._map = load_tile_map_data(
            os.path.join(os.path.dirname(__file__), 'assets', f'stage-{self.stage}.csv'),
            TILE_SIZE,
            self._tileset,
            SCALE
        )

        self._link_sprites = Group()
        self._tile_sprites = Group()
        self._enemy_sprites = Group()

        self._link: Link | None = None
        self._stair: Stair | None = None

        self._grid = dict()

        self.width = self._map.width
        self.height = self._map.height

        self.observation_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(len(self._map.tiles), 4), dtype=np.int32
        )
        self.action_space = gym.spaces.Discrete(len(Actions))

    def render(self):
        dt = 1.0 / self.metadata["render_fps"]

        if self._screen is None:
            pygame.init()
            if self.render_mode == "human":
                pygame.display.init()
                self._screen = pygame.display.set_mode((self.width, self.height + STATUS_HEIGHT))
                pygame.display.set_caption(self.spec.id)
            else:
                self._screen = pygame.Surface((self.width, self.width + STATUS_HEIGHT), pygame.SRCALPHA)
            self._game_surface = self._screen.subsurface((0, 0, self.width, self.height))
            self._status_surface = self._screen.subsurface((0, self.height, self.width, STATUS_HEIGHT))

        self._game_surface.fill(Color.GREY)

        self._tile_sprites.update(dt)
        self._enemy_sprites.update(dt)
        self._link_sprites.update(dt)

        self._tile_sprites.draw(self._game_surface)
        self._enemy_sprites.draw(self._game_surface)
        self._link_sprites.draw(self._game_surface)

        if self._game_state == 'cleared':
            self._draw_status(f"Game Clear! Steps: {self._steps} / Dist.: {self.distance_}", Color.BLUE)
        elif self._game_state == 'game_over':
            self._draw_status(f"Game Over! Steps: {self._steps} / Dist.: {self.distance_}", Color.RED)
        else:
            self._draw_status(f"Steps: {self._steps} / Dist.: {self.distance_}", Color.BLACK)

        if self.render_mode == "human":
            pygame.event.pump()
            self._clock.tick(self.metadata['render_fps'])
            pygame.display.flip()
            return None
        else:
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self._screen)), axes=(1, 0, 2)
            )

    def get_frame(self):
        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self._game_surface)), axes=(1, 0, 2)
        )

    def reset(self, **kwargs):
        super().reset(**kwargs)

        self._link_sprites.empty()
        self._tile_sprites.empty()
        self._enemy_sprites.empty()

        self._link = None
        self._stair = None
        self._grid.clear()

        self._build_stage()
        self._game_state = 'playing'

        self._steps = 0

        self.render()
        return self._generate_obs(), dict()

    def step(self, action):
        if action is None:
            self.render()
            return self._generate_obs(), 0.0, False, False, {'step': self._steps}

        front_obj = self._grid.get(self._link.front_position_)
        cur_obj = self._grid.get(self._link.position_)

        if action == Actions.move_forward:
            self._link.move_forward(front_obj)
        elif action == Actions.turn_left:
            self._link.turn_left()
        elif action == Actions.turn_right:
            self._link.turn_right()
        elif action == Actions.pick_up:
            sword = self._link.pick_up(cur_obj)
            if sword is not None:
                self._tile_sprites.remove(sword)
                self._grid[sword.position_] = None
        elif action == Actions.drop:
            sword = self._link.drop(cur_obj)
            if sword is not None:
                self._tile_sprites.add(sword)
                self._grid[sword.position_] = sword
        elif action == Actions.attack:
            enemy = self._link.attack(front_obj)
            if enemy is not None:
                self._enemy_sprites.remove(enemy)
                self._grid[enemy.position_] = None

        self._steps += 1

        self._link_sprites.user_tick(self._link)
        self._tile_sprites.user_tick(self._link)
        self._enemy_sprites.user_tick(self._link)

        if self._steps >= self.max_steps or self._link.status_ == LinkStatus.dead:
            self._game_state = 'game_over'
        elif self._link.status_ == LinkStatus.cleared:
            self._game_state = 'cleared'

        self.render()

        if self._game_state == 'playing':
            return self._generate_obs(), 0.0, False, False, self._generate_info()
        else:
            return self._generate_obs(), 0.0, False, True, self._generate_info()

    def close(self):
        if self._screen is not None:
            pygame.display.quit()

        pygame.quit()

    @property
    def distance_(self):
        return (abs(self._link.position_[0] - self._stair.position_[0])
                + abs(self._link.position_[1] - self._stair.position_[1]))

    @property
    def steps_(self):
        return self._steps

    def _build_stage(self):
        for tile in self._map.tiles:
            sprite = None
            base_obj = TILESET_BASE_TILES.get(tile.identifier)
            if base_obj is not None:
                if base_obj == Objects.water:
                    sprite = Water(tile.sprite, (tile.grid_x, tile.grid_y))
                    self._tile_sprites.add(sprite)
                elif base_obj == Objects.fire:
                    sprite = Fire(tile.sprite, (tile.grid_x, tile.grid_y))
                    self._tile_sprites.add(sprite)
                elif base_obj == Objects.stair:
                    sprite = Stair(tile.sprite, (tile.grid_x, tile.grid_y))
                    self._tile_sprites.add(sprite)
                    self._stair = sprite
                elif base_obj == Objects.wall:
                    sprite = Wall(tile.sprite, (tile.grid_x, tile.grid_y))
                    self._tile_sprites.add(sprite)
                elif base_obj == Objects.cloud:
                    sprite = Cloud(self._sprite, (tile.grid_x, tile.grid_y))
                    self._tile_sprites.add(sprite)
                elif base_obj == Objects.link:
                    self._link = Link(self._sprite, (tile.grid_x, tile.grid_y))
                    self._link_sprites.add(self._link)
            else:
                color_id, obj_id = tile.identifier // TILESET_WIDTH, tile.identifier % TILESET_WIDTH
                color, obj = TILESET_ROW_TO_COLOR[color_id], TILESET_OBJECTS[obj_id]
                if obj == Objects.sword:
                    sprite = Sword(tile.sprite, (tile.grid_x, tile.grid_y), color)
                    self._tile_sprites.add(sprite)
                else:
                    inst = self.enemy_to_inst.get(obj)
                    if inst is not None:
                        sprite = inst(self._sprite, (tile.grid_x, tile.grid_y), color)
                        self._enemy_sprites.add(sprite)

            if sprite is not None:
                self._grid[(tile.grid_x, tile.grid_y)] = sprite

    def _generate_obs(self):
        tiles = [(
            self._link.position_[0],
            self._link.position_[1],
            Objects.link.value,
            self._link.sword_.color_ if self._link.sword_ is not None else 0
        )]

        for (x, y), obj in self._grid.items():
            if isinstance(obj, Wall):
                tiles.append((x, y, Objects.wall, 0))
            elif isinstance(obj, Water):
                tiles.append((x, y, Objects.water, 0))
            elif isinstance(obj, Fire):
                tiles.append((x, y, Objects.fire, 0))
            elif isinstance(obj, Stair):
                tiles.append((x, y, Objects.stair, 0))
            elif isinstance(obj, Cloud):
                tiles.append((x, y, Objects.cloud, obj.status_))
            elif isinstance(obj, Sword):
                tiles.append((x, y, Objects.sword, obj.color_))
            elif isinstance(obj, Enemy):
                tiles.append((x, y, obj.obj, obj.color_))

        tiles = np.array(tiles)
        return tiles

    def _generate_info(self):
        return dict(
            steps=self._steps,
            distance=self.distance_,
        )

    def _draw_status(
            self,
            text: str,
            color: Tuple[int, int, int],
    ):
        font = pygame.font.SysFont(pygame.font.get_default_font(), 30)
        text = font.render(text, True, color)
        surface_width, surface_height = self._status_surface.get_size()
        text_width, text_height = text.get_size()
        self._status_surface.fill(Color.WHITE)
        self._status_surface.blit(text, (10, ((surface_height - text_height) // 2)))




import os
from .env import AlkkagiEnv
from .wrappers import RGBImgObsWrapper
from ..util import play_bgm, ObsType


_BGM_PATH = os.path.join(os.path.dirname(__file__), 'assets', 'bgm.ogg')

def _create_env(
        n_stones: int,
        n_obstacles: int,
        bgm: bool = False,
        obs_type: ObsType = 'custom',
        **kwargs
):
    if bgm:
        play_bgm(_BGM_PATH)

    env = AlkkagiEnv(
        n_stones=n_stones,
        n_obstacles=n_obstacles,
        **kwargs
    )

    if obs_type == 'image':
        env = RGBImgObsWrapper(env)

    return env


from abc import ABC, abstractmethod
import pygame
import gymnasium as gym
from .util import get_logger, wait_for_close


class ManualPlayWrapper(ABC):
    def __init__(
            self,
            env_id: str,
            debug: bool = False,
            **kwargs,
    ) -> None:
        kwargs = kwargs or {}
        kwargs['render_mode'] = 'human'
        env = gym.make(env_id, **kwargs)

        if env.unwrapped.render_mode != 'human':
            raise ValueError('"render_mode" should be "human" for the manual play.')

        self.env = env
        self._logger = get_logger(env_id, debug)

    @abstractmethod
    def handle_events(self, event: pygame.event.Event):
        raise NotImplementedError()

    @property
    def default_action_(self):
        return None

    def play(self, max_play: int = 0):
        done, steps, action = True, 0, None
        play_count = 0
        running = True

        while running:
            if done:
                play_count += 1

                if 0 < max_play < play_count:
                    break

                done, steps, action = False, 0, None
                obs, info = self.env.reset()
                self._logger.info(f'{play_count}th Play: {self.env.spec.id}')
                self._logger.info(f'Environment reset!')

                self._logger.debug(f'{steps}th Observation: {obs}')
                self._logger.debug(f'{steps}th Info: {info}')
            elif action is not None:
                self._logger.debug(f'{steps}th Action: {action or self.default_action_}')
                obs, reward, terminated, truncated, info = self.env.step(action)
                done = terminated or truncated

                steps += 1
                self._logger.debug(f'{steps}th Observation: {obs}')
                self._logger.debug(f'{steps}th Info: {info}')

                if done:
                    self._logger.info('Completed!')
                action = None
            else:
                self.env.render()

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False
                elif event.type == pygame.KEYUP and event.key == pygame.K_ESCAPE:
                    running = False
                elif event.type == pygame.KEYUP and event.key == pygame.K_BACKSPACE:
                    done = True
                else:
                    action = self.handle_events(event)
                    break

            if action is None:
                action = self.default_action_

        wait_for_close(self.env)

        self._logger.info(f'Playable environment closed!')


import os
import gymnasium as gym
from .env import ZeldaAdventureEnv
from .wrappers import RGBImgObsWrapper
from ..util import play_bgm, ObsType


_BGM_PATH = os.path.join(os.path.dirname(__file__), 'assets', 'bgm.ogg')


def _create_env(
        max_steps: int,
        stage: int,
        bgm: bool = False,
        obs_type: ObsType = 'custom',
        **kwargs
) -> gym.Env:
    if bgm:
        play_bgm(_BGM_PATH)

    env = ZeldaAdventureEnv(
        max_steps=max_steps,
        stage=stage,
        **kwargs
    )

    if obs_type == 'image':
        env = RGBImgObsWrapper(env)
    
    return env


import pygame
import numpy as np
from dataclasses import dataclass
from itertools import product
from typing import List
from .util import load_sprite


@dataclass
class TileData:
    identifier: int
    grid_x: int
    grid_y: int
    x: int
    y: int
    sprite: pygame.Surface


@dataclass
class TiledMapData:
    tiles: List[TileData]
    grid_width: int
    grid_height: int
    width: int
    height: int


def load_tile_map_data(
        path_map: str, tile_size: int, tile_set: pygame.Surface , scale: float
):
    scaled_tile_size = tile_size * scale
    tile_set_width = tile_set.get_width() // scaled_tile_size
    tile_map = np.loadtxt(path_map, delimiter=',').T
    map_width, map_height = tile_map.shape
    scaled_width, scaled_height = int(map_width * scaled_tile_size), int(map_height * scaled_tile_size)

    tiles = list()
    for x, y in product(range(map_width), range(map_height)):
        identifier = tile_map[x, y]
        if identifier > -1:
            offset = identifier % tile_set_width * tile_size, identifier // tile_set_width * tile_size
            scaled_x, scaled_y = x * scaled_tile_size, y * scaled_tile_size
            sprite = load_sprite(tile_set, (tile_size, tile_size), offset, scale=scale)
            tiles.append(
                TileData(
                    identifier, x, y, int(scaled_x), int(scaled_y), sprite
                )
            )

    return TiledMapData(tiles, map_width, map_height, scaled_width, scaled_height)


class Group(pygame.sprite.Group):
    def frame_tick(self, *args, **kwargs):
        for sprite in self.sprites():
            sprite.frame_tick(*args, **kwargs)

    def user_tick(self, *args, **kwargs):
        for sprite in self.sprites():
            sprite.user_tick(*args, **kwargs)


class Sprite(pygame.sprite.Sprite):
    rect: pygame.Rect
    image: pygame.Surface

    def frame_tick(self, *args, **kwargs):
        pass

    def user_tick(self, *args, **kwargs):
        pass

    @property
    def bb_(self):
        return self.rect.left, self.rect.top, self.rect.right, self.rect.bottom


class TiledSprite(Sprite):
    scale: float
    tile_size: int

    def set_rect(self, **kwargs):
        kwargs = {
            k: v * self.scale * self.tile_size for k, v in kwargs.items()
        }
        self.rect = self.image.get_rect(**kwargs)


import queue
import threading
import time
from typing import List, Any
import pygame
from abc import ABC, abstractmethod
import gymnasium as gym
import rpyc
from .agent import Agent
from .util import wait_for_close, get_logger
from .errors import NotAllowedUserIdError, InvalidActionError, ObservationError, EnvironmentClosedError


rpyc.core.vinegar._generic_exceptions_cache["kymnasium.errors.NotAllowedUserIdError"] = NotAllowedUserIdError
rpyc.core.vinegar._generic_exceptions_cache["kymnasium.errors.InvalidActionError"] = InvalidActionError


def evaluate(env_id: str, agent: Agent, debug: bool = False, **kwargs):
    kwargs = kwargs or {}
    kwargs['render_mode'] = 'human'
    logger = get_logger(env_id, debug)

    env = gym.make(env_id, **kwargs)
    logger.info(f'Environment is generated.')

    if env.unwrapped.render_mode != 'human':
        raise ValueError('"render_mode" should be "human" for the evaluation.')

    done, steps = False, 0
    observation, info = env.reset()
    logger.info(f'Environment is reset.')

    logger.debug(f'{steps}th Observation: {observation}')
    logger.debug(f'{steps}th Info: {info}')

    while not done:
        action = agent.act(observation, info)

        observation, reward, terminated, truncated, info = env.step(action)
        done = terminated or truncated

        logger.debug(f'{steps}th Action: {action}')

        steps += 1
        logger.debug(f'{steps}th Observation: {observation}')
        logger.debug(f'{steps}th Info: {info}')

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                done = True
            elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
                done = True

    logger.info('Completed. Press ESC key or click the exit button to exit.')
    wait_for_close(env)

    logger.info('Environment is closed.')


class RemoteEnvWrapper(ABC, rpyc.Service):
    def __init__(
            self,
            env_id: str,
            allowed_ids: List[str] = None,
            debug: bool = False,
            **kwargs
    ) -> None:
        kwargs = kwargs or {}
        kwargs['bgm'] = True

        env = gym.make(env_id, **kwargs)

        if env.unwrapped.render_mode != 'human':
            raise ValueError('"render_mode" should be "human" for the evaluation.')

        self.env = env
        self.allowed_ids = allowed_ids
        self._logger = get_logger(env_id, debug)

        self._observation = None
        self._reward = None
        self._info = {}
        self._terminated = False
        self._truncated = False

        self._action_queue = queue.Queue()
        self._result_queue = queue.Queue()
        self._main_thread_event = threading.Event()
        self._n_conns = 0

    def on_connect(self, conn: rpyc.Connection):
        connid = conn._config['connid']
        self._n_conns += 1
        self._logger.info(f"Client connected: {connid} / # connected clients: {self._n_conns}")

    def on_disconnect(self, conn):
        connid = conn._config['connid']
        self._n_conns -= 1
        self._logger.info(f"Client disconnected: {connid} / # connected clients: {self._n_conns}")

    def run(self, host: str, port: int):
        self._logger.info(f'Server initializing...')

        server = rpyc.ThreadedServer(
            self,
            hostname=host,
            port=18861,
            protocol_config={
                'import_custom_exceptions': True,
                'allow_pickle': True
            }
        )
        server_thread = threading.Thread(target=server.start, daemon=True)
        server_thread.start()

        self._logger.info(f'Server started: {host}:{port}')

        self._logger.info(f'Environment initializing...')

        self._observation, self._info = self.env.reset()
        self._reward = None
        self._terminated = False
        self._truncated = False

        self._logger.info(f'Environment reset!')

        done, steps = False, 0

        self._logger.debug(f'{steps}th Observation: {self._observation}')
        self._logger.debug(f'{steps}th Info: {self._info}')

        while not done:
            if self._main_thread_event.wait(timeout=1):
                self._main_thread_event.clear()

                try:
                    action = self._action_queue.get(timeout=10.0)
                except queue.Empty:
                    action = None

                if action is not None:
                    result = self.env.step(action)
                    self._logger.debug(f'{steps}th Action: {action}')
                    self._result_queue.put(result)
                    steps += 1
                    _, _, terminated, truncated, _ = result
                    done = terminated or truncated
                    self._logger.debug(f'{steps}th Observation: {result[0]}')
                    self._logger.debug(f'{steps}th Info: {result[-1]}')

            if self._observation is not None:
                self.env.render()

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    done = True
                elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
                    done = True

        self._logger.info('Completed!')
        self._logger.info('Wait for clients to be disconnected...')

        while self._n_conns > 0:
            time.sleep(3)

        self._logger.info('All clients are disconnected.')

        if server.active:
            server.close()

        if server_thread.is_alive():
            server_thread.join()

        self._logger.info(f'Server stopped.')

        self._logger.info('Press ESC key or click the exit button to exit.')

        wait_for_close(self.env)
        self._logger.info(f'Environment closed')

    def exposed_latest_observation(self, eval_id):
        if self.allowed_ids is not None and eval_id not in self.allowed_ids:
            self._logger.info(f'Invalid evaluation id: {eval_id}')
            raise NotAllowedUserIdError(eval_id)

        return self.serialize(self._observation), self._reward, self._terminated, self._truncated, self._info

    def exposed_step(self, eval_id, action):
        self._logger.debug(f"Step requested for {eval_id}'s action: {action}")

        if not self.verify_action(action):
            self._logger.debug(f'Invalid action: {action}')
            raise InvalidActionError(action)
        elif self.allowed_ids is not None and eval_id not in self.allowed_ids:
            self._logger.info(f'Invalid evaluation id: {eval_id}')
            raise NotAllowedUserIdError(eval_id)
        else:
            try:
                self._action_queue.put(action)
                self._main_thread_event.set()

                observation, reward, terminated, truncated, info = self._result_queue.get(timeout=10.0)

                self._observation = observation
                self._reward = reward
                self._terminated = terminated
                self._truncated = truncated
                self._info = info

                return
            except Exception as e:
                self._logger.error(f"Error occurred during step: {e} / {eval_id}'s action: {action}")
                raise e

    @abstractmethod
    def verify_action(self, action) -> bool:
        pass

    @abstractmethod
    def serialize(self, observation: Any):
        pass


def evaluate_remote(
        user_id: str,
        agent: str | Agent,
        host: str,
        port: int,
        delay: int = 3,
        debug: bool = False
):
    logger = get_logger(agent.__class__.__name__, debug)
    conn, service = None, None
    done = False
    step = 0

    while not done:
        try:
            if conn is None or conn.closed:
                conn = rpyc.connect(host, port, config={
                    'connid': user_id,
                    'allow_pickle': True
                })
                service = conn.root
                logger.info(f'Connected to the server: {host}:{port}')

            observation, reward, terminated, truncated, info = service.latest_observation(user_id)
            done = terminated or truncated

            if observation is None:
                raise ObservationError(f'Failed to obtain observation; try again later.')
            elif done:
                raise EnvironmentClosedError(f'Environment is already closed; it will stop')
            else:
                action = agent.act(observation, info)
                logger.debug(f'{step}th Observation: {observation}')
                logger.debug(f'{step}th Info: {info}')
                logger.debug(f'{step}th Action: {action}')

            if action is not None:
                service.step(user_id, action)
                logger.debug(f'{step}th interaction completed!')
                step += 1
        except ObservationError as e:
            logger.error(e)
        except InvalidActionError:
            logger.debug(f'Failed due to the invalid action; try it later')
        except NotAllowedUserIdError:
            logger.error(f'{user_id} is not allowed; Stop and check id again.')
            break
        except EnvironmentClosedError:
            logger.info('Environment is closed.')
            break
        except ConnectionError:
            logger.error(f'Failed to connect or communicate with the server: {host}:{port}')
        except Exception as e:
            logger.error(f'Failed due to unknown reason: {e}')
        finally:
            time.sleep(delay)

    logger.info('Completed!')

    if conn is not None:
        conn.close()
        logger.info(f'Disconnected to the server: {host}:{port}')






import numpy as np
import gymnasium as gym
import pygame
from minigrid.minigrid_env import MiniGridEnv
from minigrid.wrappers import RGBImgObsWrapper as MiniGridRGBImgObsWrapper, \
    FullyObsWrapper as MiniGridFullyObsWrapper
from itertools import product
from .env import Actions
from ..manual import ManualPlayWrapper


class FullyObsWrapper(MiniGridFullyObsWrapper):
    """
    (width, height, 3) numpy array observation wrapper for Grid Adventure,
    where each pixel is an encoding of the object type, color, and state.
    object type: 1=empty, 2=wall, 4=door, 5=key, 6=ball, 8=goal, 9=lava, 10=agent
    color: 0=red, 1=green, 2=blue, 3=purple, 4=yellow, 5=gray only for a door, a key, and a ball.
    state: 0=open, 1=closed, 2=locked for a door; 0=right, 1=down, 2=left, 3=up for an agent
    """
    def __init__(self, env: gym.Env):
        super().__init__(env)
        env_unwrapped = env.unwrapped

        assert isinstance(env_unwrapped, MiniGridEnv)
        width, height = env_unwrapped.width, env_unwrapped.height

        self.observation_space = gym.spaces.Box(
            low=0,
            high=1600,
            shape=(width, height, ),
            dtype=np.uint16,
        )

    def observation(self, obs):
        obs = super().observation(obs)
        img = obs['image']
        row, col = img.shape[:2]
        arr = np.zeros((row, col), dtype=np.uint16)
        for i, j in product(range(row), range(col)):
            arr[j, i] = int(''.join(map(str, img[i, j])))
        return arr


class RGBImgObsWrapper(MiniGridRGBImgObsWrapper):
    """
    Wrapper to use fully observable RGB image as observation,
    This can be used to have the agent to solve the gridworld in pixel space.
    """
    def __init__(self, env: MiniGridEnv):
        super().__init__(env)

    def observation(self, obs):
        obs = super().observation(obs)
        return obs['image']


class GridAdventureManualPlayWrapper(ManualPlayWrapper):
    KEY_TO_ACTION = {
        pygame.K_LEFT: Actions.left,
        pygame.K_RIGHT: Actions.right,
        pygame.K_UP: Actions.forward,
        pygame.K_SPACE: Actions.unlock,
        pygame.K_TAB: Actions.pickup,
        pygame.K_LSHIFT: Actions.drop,
    }

    def handle_events(self, event):
        if event.type == pygame.KEYDOWN:
            return self.KEY_TO_ACTION.get(event.key)
        return None





import os
import numpy as np
import gymnasium as gym
from .env import GridWorldEnv
from .wrappers import FullyObsWrapper, RGBImgObsWrapper
from ..util import play_bgm, ObsType


_BGM_PATH = os.path.join(os.path.dirname(__file__), 'assets', 'bgm.ogg')

def _create_env(
        max_steps: int,
        blueprint: str,
        bgm: bool = False,
        obs_type: ObsType = 'custom',
        **kwargs
) -> gym.Env:
    if bgm:
        play_bgm(_BGM_PATH)

    blueprint = np.loadtxt(blueprint, dtype=str, delimiter=',').T
    env = GridWorldEnv(max_steps=max_steps, blueprint=blueprint, **kwargs)

    if obs_type == 'image':
        env = RGBImgObsWrapper(env)
    else:
        env = FullyObsWrapper(env)

    return env




import os
from enum import IntEnum

FPS = 60
SCALE = 3.0
TILE_SIZE = 16
GAME_TILE_SIZE = int(TILE_SIZE * SCALE)
STATUS_HEIGHT = 50
SPRITE_COLOR_KEY = (146, 144, 255)


class Ids(IntEnum):
    mario = 0
    star = 26
    land = 32
    hard_block = 33
    brick_block = 34
    empty_block = 35


class Actions(IntEnum):
    noop = 0
    left = 1
    right = 2
    jump = 3


MARIO_IDLE_THRESHOLD = 10.0
MARIO_GROUND_MAX_SPEED = 750.0
MARIO_AIR_MAX_SPEED = 500.0
MARIO_MIN_ANIM_INTERVAL, MARIO_MAX_ANIM_INTERVAL = 0.005, 0.1
MARIO_INVINCIBLE_TIME = 10.0
MARIO_INVINCIBLE_ANIM_INTERVAL = 0.05
MARIO_FRICTION = 0.90
MARIO_INVINCIBLE_FRAMES = 4
MARIO_GRAVITY = 2000.0

MARIO_DEFAULT_MOVING_FORCE = 75.0
MARIO_MAX_MOVING_FORCE = 100.0
MARIO_DEFAULT_FLOATING_FORCE = 50.0
MARIO_MAX_FLOATING_FORCE = 75.0
MARIO_JUMPING_FORCE = 1000.0
MARIO_HOP_FORCE = 500.0

MARIO_COLOR_SKIN = 234, 158, 34
MARIO_COLOR_SHIRT = 181, 49, 32
MARIO_COLOR_INNER = 107, 109, 0

MARIO_COLOR_SKIN_LIGHT = 254, 204, 197
MARIO_COLOR_SKIN_WHITE = 255, 254, 255

MARIO_COLOR_SHIRT_GREEN = 12, 147, 0
MARIO_COLOR_SHIRT_RED = 181, 49, 32
MARIO_COLOR_SHIRT_BLACK = 0, 0, 0

MARIO_COLOR_INNER_ORANGE = 234, 158, 34
MARIO_COLOR_INNER_BROWN = 153, 78, 0

BULLET_BILL_MIN_SPEED = 200.0
BULLET_BILL_MAX_SPEED = 300.0
BULLET_BILL_FALLEN_SPEED = 700.0
STAR_ANIM_INTERVAL = 0.05

PATH_SPRITE = os.path.join(os.path.dirname(__file__), 'assets', 'sprite.png')
FILE_OBJECT = '_object.csv'
FILE_BACKGROUND = '_background.csv'

OFFSET_MARIO_STAND = 0, 0
OFFSET_MARIO_WALK = [(16, 0), (32, 0), (48, 0)]
OFFSET_MARIO_JUMP = 64, 0
OFFSET_MARIO_DEAD = 160, 0

OFFSET_BULLET_BILL = 96, 16

OFFSET_STAR = [(0, 32), (16, 32), (32, 32), (48, 32)]






import pygame
from typing import Tuple
from .consts import *
from ..util import load_sprite, swap_colors
from ..sprite import TiledSprite


class ZeldaSprite(TiledSprite):
    scale = SCALE
    tile_size = TILE_SIZE


class Link(ZeldaSprite):
    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[int, int]
    ):
        super().__init__()

        self._status = LinkStatus.normal
        self._direction = Directions.down
        self._position = position
        self._sword: Sword | None = None

        self._idle_index = 0
        self._idle_timer = 0

        self._action_idx = 0
        self._action_timer = 0

        images = dict()
        mapper = {
            (Directions.down, LinkStatus.normal, False): (LINK_OFFSET_DOWN_IDLE, False),
            (Directions.right, LinkStatus.normal, False): (LINK_OFFSET_RIGHT_IDLE, False),
            (Directions.left, LinkStatus.normal, False): (LINK_OFFSET_RIGHT_IDLE, True),
            (Directions.up, LinkStatus.normal, False): (LINK_OFFSET_UP_IDLE, False),
            (None, LinkStatus.picking_up, False): (LINK_OFFSET_PICK_UP, False),
            (Directions.down, LinkStatus.attacking, False): (LINK_OFFSET_DOWN_ACTION, False),
            (Directions.right, LinkStatus.attacking, False): (LINK_OFFSET_RIGHT_ACTION, False),
            (Directions.left, LinkStatus.attacking, False): (LINK_OFFSET_RIGHT_ACTION, True),
            (Directions.up, LinkStatus.attacking, False): (LINK_OFFSET_UP_ACTION, False),
            (Directions.down, LinkStatus.attacking, True): (LINK_OFFSET_DOWN_ATTACK, False),
            (Directions.right, LinkStatus.attacking, True): (LINK_OFFSET_RIGHT_ATTACK, False),
            (Directions.left, LinkStatus.attacking, True): (LINK_OFFSET_RIGHT_ATTACK, True),
            (Directions.up, LinkStatus.attacking, True): (LINK_OFFSET_UP_ATTACK, False),
        }

        for (direction, status, is_swap_color), (offsets, flip_x) in mapper.items():
            frames = []

            for offset in offsets:
                if len(offset) == 2:
                    frames.append(load_sprite(sprite, (TILE_SIZE, TILE_SIZE), offset, flip_x, scale=SCALE))
                else:
                    frames.append(load_sprite(sprite, offset[2:], offset[:2], flip_x, scale=SCALE))

            if is_swap_color:
                for color, rgb in PALETTE_SWORD.items():
                    images[(direction, status, color)] = [
                        swap_colors(frame, [((255, 255, 255), rgb)]) for frame in frames
                    ]
            else:
                images[(direction, status)] = frames
        self._images = images

        self.image = self._images[(self._direction, self._status)][0]
        self.set_rect(
            x=self._position[0],
            y=self._position[1]
        )

    @property
    def direction_(self):
        return self._direction

    @property
    def status_(self):
        return self._status

    @property
    def sword_(self):
        return self._sword

    @property
    def position_(self):
        return self._position

    @property
    def front_position_(self):
        cx, cy = self._position
        dx, dy = DIRECTION_TO_VECTOR[self._direction]
        return cx + dx, cy + dy

    def _stop_action(self):
        self._status = LinkStatus.normal
        self._action_idx = 0
        self._action_timer = 0

    def turn_left(self):
        self._stop_action()
        self._direction = (self._direction + 1) % len(Directions)

    def turn_right(self):
        self._stop_action()
        self._direction = (self._direction - 1) % len(Directions)

    def move_forward(self, front_obj: pygame.sprite.Sprite | None):
        self._stop_action()
        if front_obj is None or isinstance(front_obj, (Sword, Water, Fire, Stair)):
            self._position = self.front_position_
            self.set_rect(
                x=self._position[0],
                y=self._position[1]
            )
            if isinstance(front_obj, (Water, Fire)):
                self._status = LinkStatus.dead
            elif isinstance(front_obj, Stair):
                self._status = LinkStatus.cleared
        elif isinstance(front_obj, Cloud) and front_obj.passable_:
            self._position = self.front_position_
            self.set_rect(
                x=self._position[0],
                y=self._position[1]
            )

    def pick_up(self, sword: pygame.sprite.Sprite | None):
        if not isinstance(sword, Sword):
            return None

        if self._sword is None:
            self._status = LinkStatus.picking_up
            self._sword = sword

        return self._sword

    def drop(self, cur_obj: pygame.sprite.Sprite | None):
        if cur_obj is not None:
            return None

        if self._sword is not None:
            self._stop_action()
            sword = self._sword
            sword.position_ = self._position
            self._sword = None
            return sword

        return None

    def attack(self, enemy: pygame.sprite.Sprite | None) -> 'Enemy | None':
        self._status = LinkStatus.attacking

        if not isinstance(enemy, Enemy):
            return None

        damage = LINK_ATTACK_BARE_DAMAGE
        if self._sword is not None:
            if self._sword.color_ == enemy.color_:
                damage = LINK_ATTACK_CRITICAL_DAMAGE
            else:
                damage = LINK_ATTACK_NORMAL_DAMAGE

        enemy.take_damage(damage)

        if enemy.status_ == EnemyStatus.dead:
            return enemy

        return None

    def update(self, dt: float, *args, **kwargs):
        if self._status == LinkStatus.normal:
            self._idle_timer += dt
            if self._idle_timer >= DEFAULT_ANIM_INTERVAL:
                self._idle_timer = 0
                self._idle_index = (self._idle_index + 1) % len(self._images[(self._direction, self._status)])
                self.image = self._images[(self._direction, self._status)][self._idle_index]
        else:
            self._action_timer += dt

            if self._status == LinkStatus.picking_up:
                interval = DEFAULT_ANIM_INTERVAL
                frames = self._images[(None, LinkStatus.picking_up)]
            else:
                if self._sword is not None:
                    interval = ATTACK_ANIM_INTERVAL
                    frames = self._images[(self._direction, LinkStatus.attacking, self._sword.color_)]
                else:
                    interval = DEFAULT_ANIM_INTERVAL
                    frames = self._images[(self._direction, LinkStatus.attacking)]

            self.image = frames[self._action_idx]

            if self._direction == Directions.down or self._direction == Directions.right:
                self.set_rect(x = self._position[0], y = self._position[1])
            elif self._direction == Directions.left:
                self.set_rect(right=self._position[0] + 1, y=self._position[1])
            elif self._direction == Directions.up:
                self.set_rect(x=self._position[0], bottom=self._position[1] + 1)

            if self._action_timer >= interval:
                self._action_timer = 0
                self._action_idx += 1

            if self._action_idx >= len(frames):
                self._stop_action()

class Wall(ZeldaSprite):
    def __init__(
            self,
            image: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()
        self.image = image
        self.set_rect(x=position[0], y=position[1])


class Water(ZeldaSprite):
    def __init__(
            self,
            image: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()
        self.image = image
        self.set_rect(x=position[0], y=position[1])

class Fire(ZeldaSprite):
    def __init__(
            self,
            image: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()
        self.image = image
        self.set_rect(x=position[0], y=position[1])


class Stair(ZeldaSprite):
    def __init__(
            self,
            image: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()
        self._position = position
        self.image = image
        self.set_rect(x=position[0], y=position[1])

    @property
    def position_(self):
        return self._position


class Cloud(ZeldaSprite):
    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()

        size = TILE_SIZE, TILE_SIZE
        self._position = position
        self._counter = 0

        self._images = [
            load_sprite(sprite, size, offset, False, scale=SCALE)
            for offset in CLOUD_OFFSET
        ]
        self._status = CloudStatus.disappeared
        self.image = self._images[0]
        self.set_rect(x=position[0], y=position[1])

    @property
    def status_(self):
        return self._status

    @property
    def passable_(self):
        return self._status in (CloudStatus.disappeared, CloudStatus.disappeared_again)

    def user_tick(self, cur_obj: pygame.sprite.Sprite | None):
        if isinstance(cur_obj, Link) and cur_obj.position_ == self._position:
            return

        self._counter += 1
        if self._counter % CLOUD_UNIT_DURATION == 0:
            self._status = (self._status + 1) % len(CloudStatus)

    def update(self, *args, **kwargs):
        if self._status == CloudStatus.disappeared or self._status == CloudStatus.disappeared_again:
            self.image.set_alpha(0)
        else:
            if self._status == CloudStatus.appearing_slightly or self._status == CloudStatus.disappearing_more:
                self.image = self._images[0]
            elif self._status == CloudStatus.appearing_more or self._status == CloudStatus.disappearing_slightly:
                self.image = self._images[1]
            else:
                self.image = self._images[2]
            self.image.set_alpha(255)


class Sword(ZeldaSprite):
    def __init__(
            self,
            image: pygame.Surface,
            position: Tuple[int, int],
            color: int,
    ):
        super().__init__()

        self._color = color
        self._position = position
        self._is_picked_up = False

        self.image = image
        self.set_rect(x=position[0], y=position[1])

    @property
    def color_(self):
        return self._color

    @property
    def position_(self):
        return self._position

    @position_.setter
    def position_(self, position: Tuple[int, int]):
        self._position = position
        self.rect = self.image.get_rect(x=position[0] * TILE_SIZE * SCALE, y=position[1] * TILE_SIZE * SCALE)

    def update(self, *args, **kwargs):
        if self._is_picked_up:
            self.image.set_alpha(0)
        else:
            self.image.set_alpha(255)


class Enemy(ZeldaSprite):
    obj: int = Objects.none
    mapper: dict = None
    default_direction: int = None
    hit_points: float = None

    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[int, int],
            color: int,
    ):
        super().__init__()

        self._status = EnemyStatus.normal
        self._direction = self.default_direction
        self._position = position
        self._color = color
        self._hit_points = self.hit_points

        self._idle_idx = 0
        self._idle_timer = 0

        self._images = dict()

        for direction, (offsets, flip_x) in self.mapper.items():
            size = TILE_SIZE, TILE_SIZE
            frames = [
                load_sprite(sprite, size, offset, flip_x, scale=SCALE)
                for offset in offsets
            ]
            palette = list(zip(PALETTE_ENEMY[Colors.blue], PALETTE_ENEMY[self._color]))
            frames = [
                swap_colors(
                    frame, palette
                ) for frame in frames
            ]
            self._images[direction] = frames

        self.image = self._images[self._direction][self._idle_idx]
        self.set_rect(x=position[0], y=position[1])

    @property
    def status_(self):
        return self._status

    @property
    def position_(self):
        return self._position

    @property
    def direction_(self):
        return self._direction

    @property
    def hit_points_(self):
        return self._hit_points

    @property
    def color_(self):
        return self._color

    def take_damage(self, damage: float):
        self._hit_points -= damage
        if self._hit_points <= 1e-4:
            self._status = EnemyStatus.dead

    def update(self, dt: float):
        self._idle_timer += dt
        if self._idle_timer >= DEFAULT_ANIM_INTERVAL:
            self._idle_timer = 0
            self._idle_idx = (self._idle_idx + 1) % len(self._images[self._direction])
            self.image = self._images[self._direction][self._idle_idx]


class Darknut(Enemy):
    obj = Objects.darknut
    mapper = {
        Directions.down: (DARKNUT_OFFSET_DOWN, False),
        Directions.right: (DARKNUT_OFFSET_RIGHT, False),
        Directions.left: (DARKNUT_OFFSET_RIGHT, True),
        Directions.up: (DARKNUT_OFFSET_DOWN, False),
    }
    default_direction = Directions.down
    hit_points = DARKNUT_HIT_POINT


class Goriya(Enemy):
    obj = Objects.goriya
    mapper = {
        Directions.down: (GORIYA_OFFSET_DOWN, False),
        Directions.right: (GORIYA_OFFSET_RIGHT, False),
        Directions.left: (GORIYA_OFFSET_RIGHT, True),
        Directions.up: (GORIYA_OFFSET_DOWN, False),
    }
    default_direction = Directions.down
    hit_points = GORIYA_HIT_POINT


class Wizzrobe(Enemy):
    obj = Objects.wizzrobe
    mapper = {
        Directions.down: (WIZZROBE_OFFSET_DOWN, False),
        Directions.up: (WIZZROBE_OFFSET_UP, False)
    }
    default_direction = Directions.down
    hit_points = WIZZROBE_HIT_POINT


class Rope(Enemy):
    obj = Objects.rope
    mapper = {
        Directions.right: (ROPE_OFFSET_RIGHT, False),
        Directions.left: (ROPE_OFFSET_RIGHT, True)
    }
    default_direction = Directions.right
    hit_points = ROPE_HIT_POINT


class Moblin(Enemy):
    obj = Objects.moblin
    mapper = {
        Directions.down: (MOBLIN_OFFSET_DOWN, False),
        Directions.right: (MOBLIN_OFFSET_RIGHT, False),
        Directions.left: (MOBLIN_OFFSET_RIGHT, True),
        Directions.up: (MOBLIN_OFFSET_DOWN, False),
    }
    default_direction = Directions.down
    hit_points = MOBLIN_HIT_POINT


class Armos(Enemy):
    obj = Objects.armos
    mapper = {
        Directions.down: (ARMOS_OFFSET_DOWN, False),
        Directions.up: (ARMOS_OFFSET_UP, False),
    }
    default_direction = Directions.down
    hit_points = ARMOS_HIT_POINT


class Octorok(Enemy):
    obj = Objects.octorok
    mapper = {
        Directions.down: (OCTOROK_OFFSET, False),
    }
    default_direction = Directions.down
    hit_points = OCTOROK_HIT_POINT


class Keese(Enemy):
    obj = Objects.keese
    mapper = {
        Directions.down: (KEESE_OFFSET, False),
    }
    default_direction = Directions.down
    hit_points = KEESE_HIT_POINT


class Tektite(Enemy):
    obj = Objects.tektite
    mapper = {
        Directions.down: (TEKTITE_OFFSET, False),
    }
    default_direction = Directions.down
    hit_points = TEKTITE_HIT_POINT



import random
from typing import Tuple, List, Literal
import numpy as np
import pygame
import pygame.freetype
import gymnasium as gym
from .objs import Mario, BulletBill, Star, WorldObjects
from .consts import *
from ..color import Color
from ..sprite import load_tile_map_data, Group


class BulletBillEnv(gym.Env):
    metadata = {
        'render_modes': ['human', 'rgb_array'],
        'render_fps': FPS
    }

    def __init__(
            self,
            game_duration: float,
            init_spawn_interval: float,
            min_spawn_interval: float,
            max_spawns: int,
            stage: int = 1,
            continuous_action: bool = False,
            max_spawn_duration: float = None,
            frame_skip: int = 4,
            render_mode=None
    ):
        self.render_mode = render_mode
        self.stage = stage
        self.continuous_action = continuous_action
        self.max_spawn_duration = min(max_spawn_duration, game_duration) if max_spawn_duration else game_duration
        self.frame_skip = frame_skip
        self.game_duration = game_duration
        self.init_spawn_interval = init_spawn_interval
        self.min_spawn_interval = min_spawn_interval
        self.max_spawns = max_spawns

        self._time_elapsed = 0
        self._timer = 0
        self._spawn_interval = init_spawn_interval
        self._game_state: Literal['init', 'playing', 'game_over', 'cleared'] = 'init'

        sprite = pygame.image.load(PATH_SPRITE)
        sprite.set_colorkey(SPRITE_COLOR_KEY)
        sprite = pygame.transform.scale(
            sprite,
            (sprite.get_width() * SCALE, sprite.get_height() * SCALE)
        )
        self._sprite = sprite

        self._screen = None
        self._game_surface = None
        self._status_surface = None
        self._clock = pygame.time.Clock()

        self._background_map = load_tile_map_data(
            os.path.join(os.path.dirname(__file__), 'assets', f'stage-{self.stage}{FILE_BACKGROUND}'),
            TILE_SIZE,
            self._sprite,
            SCALE
        )

        self._object_map = load_tile_map_data(
            os.path.join(os.path.dirname(__file__), 'assets', f'stage-{self.stage}{FILE_OBJECT}'),
            TILE_SIZE,
            self._sprite,
            SCALE
        )

        self.width, self.height = self._background_map.width, self._background_map.height

        self._character_sprites = Group()
        self._background_sprites = Group()
        self._static_sprites = Group()

        self._mario: Mario | None = None
        self._bullet_bills: List[BulletBill] = []
        self._stars: List[Star] = []
        self._platforms: List[WorldObjects] = []

        self._n_star = sum([tile.identifier == Ids.star for tile in self._object_map.tiles])

        self.observation_space = gym.spaces.Dict({
            "mario": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(8,)),
            "bullet_bills": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(self.max_spawns, 5)),
            "stars": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(self._n_star, 4)),
        })
        if self.continuous_action:
            self.action_space = gym.spaces.Box(low=-np.inf, high=np.inf, shape=(2,))
        else:
            self.action_space = gym.spaces.Discrete(4)

    def render(self):
        fps = self.metadata["render_fps"]
        dt = 1.0 / fps

        if self._screen is None:
            pygame.init()
            if self.render_mode == "human":
                pygame.display.init()
                self._screen = pygame.display.set_mode((self.width, self.height + STATUS_HEIGHT))
                pygame.display.set_caption(self.spec.id)
            else:
                self._screen = pygame.Surface((self.width, self.width + STATUS_HEIGHT), pygame.SRCALPHA)
            self._game_surface = self._screen.subsurface((0, 0, self.width, self.height))
            self._status_surface = self._screen.subsurface((0, self.height, self.width, STATUS_HEIGHT))

        self._game_surface.fill(Color.SKY_BLUE)

        self._character_sprites.update(dt)

        self._background_sprites.draw(self._game_surface)
        self._static_sprites.draw(self._game_surface)
        self._character_sprites.draw(self._game_surface)

        if self._game_state == 'cleared':
            self._draw_status("Game Clear!", Color.BLUE)
        elif self._game_state == 'game_over':
            self._draw_status(f"Game Over! You survived for {self._time_elapsed:.2f} seconds", Color.RED)
        else:
            self._draw_status(f"Time elapsed: {self._time_elapsed:.2f} seconds", Color.BLACK)

        if self.render_mode == "human":
            pygame.event.pump()
            self._clock.tick(self.metadata['render_fps'])
            pygame.display.flip()
            return None
        else:
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self._screen)), axes=(1, 0, 2)
            )

    def get_frame(self):
        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self._game_surface)), axes=(1, 0, 2)
        )

    def reset(self, **kwargs):
        super().reset(**kwargs)

        self._character_sprites.empty()
        self._background_sprites.empty()
        self._static_sprites.empty()

        self._mario = None
        self._bullet_bills.clear()
        self._stars.clear()

        self._time_elapsed = 0
        self._timer = 0
        self._spawn_interval = self.init_spawn_interval

        self._build_stage()
        self._game_state = 'playing'

        self.render()

        return self._generate_obs(), dict()

    def step(self, action):
        obs, reward, terminated, truncated, info = None, 0.0, False, False, {}

        for _ in range(self.frame_skip):
            obs, reward, terminated, truncated, info = self._step_internal(action)
            if terminated or truncated:
                break
        return obs, reward, terminated, truncated, info

    def close(self):
        if self._screen is not None:
            pygame.display.quit()

        pygame.quit()

    def _build_stage(self):
        for tile in self._background_map.tiles:
            background = WorldObjects(tile.sprite,  (tile.x, tile.y))
            self._background_sprites.add(background)

        for tile in self._object_map.tiles:
            if tile.identifier == Ids.mario:
                self._mario = Mario(
                    sprite=self._sprite,
                    position=(tile.x, tile.y)
                )
                self._character_sprites.add(self._mario)
            elif tile.identifier == Ids.star:
                star = Star(
                    self._sprite, (tile.x, tile.y)
                )
                self._character_sprites.add(star)
                self._stars.append(star)
            else:
                platform = WorldObjects(tile.sprite, (tile.x, tile.y))
                self._static_sprites.add(platform)
                self._platforms.append(platform)

    def _spawn_bullet_bill(self):
        if len(self._bullet_bills) >= self.max_spawns:
            return

        target_y = random.uniform(-GAME_TILE_SIZE * 3, GAME_TILE_SIZE * 3) + self._mario.bb_[1]
        target_y = np.clip(target_y, 0, self._mario.bb_[1])
        is_left = random.choice([True, False])
        x = -GAME_TILE_SIZE if is_left else self.width
        speed = random.uniform(BULLET_BILL_MIN_SPEED, BULLET_BILL_MAX_SPEED)
        bullet_bill = BulletBill(
            sprite=self._sprite,
            position=(x, target_y),
            velocity=speed if is_left else -speed,
            flip=is_left,
        )
        self._bullet_bills.append(bullet_bill)
        self._character_sprites.add(bullet_bill)

    def _generate_obs(self):
        mario = np.array((
            *self._mario.bb_, *self._mario.velocity_,
            1 if self._mario.invincible_ else 0, 1 if self._mario.jumping_ else 0
        ), dtype=np.float32)

        bullets = np.zeros(shape=(self.max_spawns, 5), dtype=np.float32)
        bullets_active = [bullet for bullet in self._bullet_bills if bullet.state_ == 'active']

        for i, bullet in enumerate(bullets_active):
            bullets[i] = (*bullet.bb_, bullet.velocity_)

        stars = np.zeros(shape=(self._n_star, 4), dtype=np.float32)
        stars_active = [star for star in self._stars if star.active_]

        for i, star in enumerate(stars_active):
            stars[i] = (*star.bb_, )

        return {
            "mario": mario,
            "bullet_bills": bullets,
            "stars": stars,
        }

    def _draw_status(
            self,
            text: str,
            color: Tuple[int, int, int],
    ):
        font = pygame.font.SysFont(pygame.font.get_default_font(), 30)
        text = font.render(text, True, color)
        surface_width, surface_height = self._status_surface.get_size()
        text_width, text_height = text.get_size()
        self._status_surface.fill(Color.WHITE)
        self._status_surface.blit(text, (10, ((surface_height - text_height) // 2)))

    def _step_internal(self, action):
        fps = self.metadata["render_fps"]
        dt = 1.0 / fps

        self._timer += dt
        self._time_elapsed += dt

        if self._time_elapsed >= self.game_duration:
            self._game_state = 'cleared'
            self.render()
            return self._generate_obs(), 0.0, True, False, dict()

        if not self.continuous_action:
            walk_force, jump_force = 0.0, 0.0
            if action == Actions.left:
                walk_force = -MARIO_DEFAULT_MOVING_FORCE
            elif action == Actions.right:
                walk_force = MARIO_DEFAULT_MOVING_FORCE
            elif action == Actions.jump:
                jump_force = -MARIO_JUMPING_FORCE

            action = (walk_force, jump_force)

        self._mario.apply_force(action)

        if self._timer >= self._spawn_interval:
            self._timer = 0
            self._spawn_bullet_bill()

        self._spawn_interval = self.init_spawn_interval * (
                self.min_spawn_interval / self.init_spawn_interval
        ) ** (self._time_elapsed / self.max_spawn_duration)

        self._character_sprites.frame_tick(dt=dt, screen_size=(self.width, self.height), platforms=self._platforms)

        self._mario.check_star_collision(self._stars)
        self._mario.check_bullet_collision(self._bullet_bills)

        bullets_oob = [bullet for bullet in self._bullet_bills if bullet.state_ == 'oob']
        for bullet in bullets_oob:
            bullet.kill()
            self._bullet_bills.remove(bullet)

        stars_inactive = [star for star in self._stars if not star.active_]
        for star in stars_inactive:
            star.kill()
            self._stars.remove(star)

        if not self._mario.active_:
            self._game_state = 'game_over'

        self.render()

        if self._game_state == 'playing':
            return self._generate_obs(), 0.0, False, False, {"time_elapsed": self._time_elapsed}
        else:
            return self._generate_obs(), 0.0, False, True, {"time_elapsed": self._time_elapsed}


import pygame
import numpy as np
from typing import Tuple, List, Literal
from .consts import *
from ..util import load_sprite, swap_colors
from ..sprite import Sprite


class Mario(Sprite):
    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[int, int]
    ):
        super().__init__()
        self._active = True
        self._is_right = True
        self._is_grounded = False
        self._is_invincible = False

        self._walk_idx = 0
        self._walk_timer = 0

        self._invincible_anim_idx = 0
        self._invincible_anim_timer = 0
        self._invincible_timer = 0

        self._velocity = pygame.Vector2(0, 0)

        images = dict()
        mapper = {
            'right_stand': (OFFSET_MARIO_STAND, False),
            'left_stand': (OFFSET_MARIO_STAND, True),
            'right_walk': (OFFSET_MARIO_WALK, False),
            'left_walk': (OFFSET_MARIO_WALK, True),
            'right_jump': (OFFSET_MARIO_JUMP, False),
            'left_jump': (OFFSET_MARIO_JUMP, True),
            'dead': (OFFSET_MARIO_DEAD, False)
        }
        swaps = [
            # Default mario
            None,
            # Green-pale mario
            [
                (MARIO_COLOR_SKIN, MARIO_COLOR_SKIN_WHITE),
                (MARIO_COLOR_SHIRT, MARIO_COLOR_SHIRT_GREEN),
                (MARIO_COLOR_INNER, MARIO_COLOR_INNER_ORANGE),
            ],
            # Red-pale mario
            [
                (MARIO_COLOR_SKIN, MARIO_COLOR_SKIN_WHITE),
                (MARIO_COLOR_SHIRT, MARIO_COLOR_SHIRT_RED),
                (MARIO_COLOR_INNER, MARIO_COLOR_INNER_ORANGE),
            ],
            # Black-darker mario
            [
                (MARIO_COLOR_SKIN, MARIO_COLOR_SKIN_LIGHT),
                (MARIO_COLOR_SHIRT, MARIO_COLOR_SHIRT_BLACK),
                (MARIO_COLOR_INNER, MARIO_COLOR_INNER_BROWN),
            ],
        ]
        for key, (offsets, flip_x) in mapper.items():
            size = TILE_SIZE, TILE_SIZE

            if isinstance(offsets, list):
                frames = [
                    load_sprite(sprite, size, offset, flip_x, scale=SCALE)
                    for offset in offsets
                ]
                for i, swap in enumerate(swaps):
                    if swap is None:
                        images[f'{key}/{i}'] = frames
                    else:
                        images[f'{key}/{i}'] = [swap_colors(frame, swap) for frame in frames]
            else:
                frame = load_sprite(sprite, size, offsets, flip_x, scale=SCALE)
                for i, swap in enumerate(swaps):
                    if swap is None:
                        images[f'{key}/{i}'] = frame
                    else:
                        images[f'{key}/{i}'] = swap_colors(frame, swap)

        self._images = images
        self.image = self._images['right_stand/0']
        self.rect = self.image.get_rect(x=position[0], y=position[1])

    @property
    def velocity_(self):
        return self._velocity

    @property
    def invincible_(self):
        return self._is_invincible

    @property
    def active_(self):
        return self._active

    @property
    def jumping_(self):
        return not self._is_grounded

    def apply_force(self, force: Tuple[float, float]):
        fx, fy = force

        if self._is_grounded:
            self._velocity.x += np.clip(fx, -MARIO_MAX_MOVING_FORCE, MARIO_MAX_MOVING_FORCE)
        else:
            self._velocity.x += np.clip(fx, -MARIO_MAX_FLOATING_FORCE, MARIO_MAX_FLOATING_FORCE)

        if self._is_grounded and fy < 0:
            self._velocity.y = fy
            self._is_grounded = False

    def frame_tick(self, dt: float, screen_size: Tuple[int, int], platforms: List['WorldObjects'], *args, **kwargs):
        if self._is_grounded:
            self._velocity.x = np.clip(self._velocity.x, -MARIO_GROUND_MAX_SPEED, MARIO_GROUND_MAX_SPEED)
        else:
            self._velocity.x = np.clip(self._velocity.x, -MARIO_AIR_MAX_SPEED, MARIO_AIR_MAX_SPEED)
        self._velocity.y += MARIO_GRAVITY * dt

        if self._velocity.x > 0:
            self._is_right = True
        elif self._velocity.x < 0:
            self._is_right = False

        if abs(self._velocity.x) < MARIO_IDLE_THRESHOLD:
            self._velocity.x = 0.0

        self.rect.x += self._velocity.x * dt
        self.rect.left = max(0, self.rect.left)
        self.rect.right = min(screen_size[0], self.rect.right)

        self._velocity.x *= MARIO_FRICTION

        for platform in platforms:
            if self.rect.colliderect(platform.rect):
                if self._velocity.x > 0:
                    self.rect.right = platform.rect.left
                elif self._velocity.x < 0:
                    self.rect.left = platform.rect.right
        self.rect.y += (self._velocity.y * dt)
        self._is_grounded = False

        for platform in platforms:
            if self.rect.colliderect(platform.rect):
                if self._velocity.y > 0:
                    self.rect.bottom = platform.rect.top
                    self._velocity.y = 0
                    self._is_grounded = True
                elif self._velocity.y < 0:
                    self.rect.top = platform.rect.bottom
                    self._velocity.y = 0

        if self._is_invincible:
            if self._invincible_timer > 0:
                self._invincible_timer -= dt
            else:
                self._is_invincible = False
                self._invincible_timer = 0

        if self.rect.bottom > screen_size[1]:
            self._active = False

    def update(self, dt: float):
        if self._is_invincible:
            self._invincible_anim_timer += dt
            if self._invincible_anim_timer >= MARIO_INVINCIBLE_ANIM_INTERVAL:
                self._invincible_anim_timer = 0
                self._invincible_anim_idx = (self._invincible_anim_idx + 1) % MARIO_INVINCIBLE_FRAMES
        else:
            self._invincible_anim_idx = 0

        inv_idx = self._invincible_anim_idx
        if not self._active:
            self.image = self._images['dead/0']
        elif self._is_grounded:
            if abs(self._velocity.x) >= MARIO_IDLE_THRESHOLD:
                self._walk_timer += dt
                speed_factor = abs(self._velocity.x) / MARIO_GROUND_MAX_SPEED
                delay = MARIO_MAX_ANIM_INTERVAL - (MARIO_MAX_ANIM_INTERVAL - MARIO_MIN_ANIM_INTERVAL) * speed_factor
                if self._walk_timer >= delay:
                    self._walk_timer = 0
                    self._walk_idx = (self._walk_idx + 1) % len(self._images['right_walk/0'])
                self.image = self._images[f'right_walk/{inv_idx}'][self._walk_idx] if self._is_right else self._images[f'left_walk/{inv_idx}'][self._walk_idx]
            else:
                self.image = self._images[f'right_stand/{inv_idx}'] if self._is_right else self._images[
                    f'left_stand/{inv_idx}']
        elif not self._is_grounded:
            self.image = self._images[f'right_jump/{inv_idx}'] if self._is_right else self._images[f'left_jump/{inv_idx}']

    def check_platform_collision(self, platforms: List['WorldObjects']):
        self._is_grounded = False

        for platform in platforms:
            if self.rect.colliderect(platform.rect):
                if self._velocity.x > 0:
                    self.rect.right = platform.rect.left
                elif self._velocity.x < 0:
                    self.rect.left = platform.rect.right

                if self._velocity.y > 0:
                    self.rect.bottom = platform.rect.top
                    self._velocity.y = 0
                    self._is_grounded = True
                elif self._velocity.y < 0:
                    self.rect.top = platform.rect.bottom
                    self._velocity.y = 0

    def check_star_collision(self, stars: List['Star']):
        for star in stars:
            if self.rect.colliderect(star.rect):
                self._is_invincible = True
                self._invincible_timer += MARIO_INVINCIBLE_TIME
                star.active_ = False
                break

    def check_bullet_collision(self, bullets: List['BulletBill']):
        for bullet in bullets:
            if self.rect.colliderect(bullet.rect):
                if bullet.state_ == 'active':
                    if self._velocity.y > 0 and self.rect.bottom < bullet.rect.centery:
                        self._velocity.y = -MARIO_HOP_FORCE
                        bullet.state_ = 'fallen'
                    elif self._is_invincible:
                        bullet.state_ = 'fallen'
                    else:
                        self._active = False


class BulletBill(Sprite):
    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[float, float],
            velocity: float,
            flip: bool = False
    ):
        super().__init__()

        self._velocity = velocity
        self._state: Literal['active', 'oob', 'fallen'] = 'active'
        self.image = load_sprite(sprite, (TILE_SIZE, TILE_SIZE), OFFSET_BULLET_BILL, flip, scale=SCALE)
        self.rect = self.image.get_rect(x=position[0], y=position[1])

    @property
    def velocity_(self):
        return self._velocity

    @property
    def state_(self):
        return self._state

    @state_.setter
    def state_(self, state: Literal['active', 'oob', 'fallen']):
        self._state = state

    def frame_tick(self, dt: float, screen_size: Tuple[int, int], *args, **kwargs):
        if self._state == 'active':
            self.rect.centerx += self._velocity * dt
        elif self._state == 'fallen':
            self.rect.centery += BULLET_BILL_FALLEN_SPEED * dt

        if self._state == 'active':
            if self._velocity > 0 and self.rect.left > screen_size[0]:
                self._state = 'oob'
            elif self._velocity < 0 and self.rect.right < 0:
                self._state = 'oob'
        elif self._state == 'fallen':
            if self.rect.top > screen_size[1]:
                self._state = 'oob'

    def update(self, dt: float, *args, **kwargs):
        pass


class Star(Sprite):
    def __init__(
            self,
            sprite: pygame.Surface,
            position: Tuple[int, int]
    ):
        super().__init__()

        self._anim_idx = 0
        self._anim_timer = 0
        self._active = True

        self._images = [
            load_sprite(sprite, (TILE_SIZE, TILE_SIZE), offset, scale=SCALE)
            for offset in OFFSET_STAR
        ]

        self.image = self._images[0]
        self.rect = self.image.get_rect(x=position[0], y=position[1])

    @property
    def active_(self):
        return self._active

    @active_.setter
    def active_(self, active: bool):
        self._active = active

    def update(self, dt: float, *args, **kwargs):
        self._anim_timer += dt
        if self._anim_timer >= STAR_ANIM_INTERVAL:
            self._anim_timer = 0
            self._anim_idx = (self._anim_idx + 1) % len(self._images)
        self.image = self._images[self._anim_idx]

    def frame_tick(self, *args, **kwargs):
        pass


class WorldObjects(Sprite):
    def __init__(
            self,
            image: pygame.Surface,
            position: Tuple[int, int],
    ):
        super().__init__()

        self.image = image
        self.rect = self.image.get_rect(x=position[0], y=position[1])

    def update(self, *args, **kwargs):
        pass

    def frame_tick(self, *args, **kwargs):
        pass


import os
import sys
from typing import Literal, Tuple, List
import gymnasium as gym
import pygame
import logging
import numpy as np
from itertools import product
from dataclasses import dataclass


type ObsType = Literal['image', 'custom']

def wait_for_close(env: gym.Env):
    if env is None:
        return

    running = True

    while running:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
                running = False

    env.close()


def play_bgm(path: str) -> None:
    if not os.path.exists(path):
        return

    try:
        if not pygame.mixer.get_init():
            pygame.mixer.init()

        pygame.mixer.music.load(path)
        pygame.mixer.music.set_volume(0.5)
        pygame.mixer.music.play(loops=-1)
    except pygame.error as e:
        import traceback
        traceback.print_exc()
        pass


def load_sprite(
        sprite: pygame.Surface,
        size: Tuple[int, int],
        offset: Tuple[int, int],
        flip_x: bool = False,
        flip_y: bool = False,
        scale: float = 1,
) -> pygame.Surface:
    size = (size[0] * scale, size[1] * scale)
    offset = (offset[0] * scale, offset[1] * scale)
    img = sprite.subsurface(*offset, *size)
    img = pygame.transform.flip(img, flip_x, flip_y)

    return img


def swap_colors(
        surface: pygame.Surface,
        colors: List[Tuple[Tuple[int, int, int], Tuple[int, int, int]]]
):
    copy = surface.copy()

    for old, new in colors:
        mask = pygame.mask.from_threshold(copy, old, threshold=(1, 1, 1, 255))
        masked_surface = mask.to_surface(setcolor=new, unsetcolor=(0, 0, 0, 0))
        copy.blit(masked_surface, (0, 0))

    return copy


def get_logger(name: str, debug: bool = False):
    logger = logging.getLogger(name)

    if not logger.hasHandlers():
        formatter = logging.Formatter('[%(levelname)s - %(asctime)s - %(name)s] %(message)s')

        normal_handler = logging.StreamHandler(sys.stdout)
        normal_handler.setFormatter(formatter)
        normal_handler.setLevel(logging.DEBUG)

        error_handler = logging.StreamHandler(sys.stderr)
        error_handler.setFormatter(formatter)
        error_handler.setLevel(logging.WARNING)

        logger.addHandler(normal_handler)
        logger.addHandler(error_handler)

    logger.setLevel(
        logging.DEBUG if debug else logging.INFO
    )
    return logger



import gymnasium as gym
import numpy as np
from typing import Any
import pygame
import pygame.math
import math
from .objs import SlingShot
from .env import AlkkagiEnv
from ..evaluate import RemoteEnvWrapper
from ..agent import Agent
from ..manual import ManualPlayWrapper


class AlkkagiManualPlayWrapper(ManualPlayWrapper):
    def __init__(self, env_id: str, debug: bool = False, agent: Agent = None, agent_turn: int = None, **kwargs):
        super().__init__(env_id, debug, **kwargs)

        self._is_dragging = False
        self._pos_start = None
        self._pos_end = None
        self._selected_stone = None
        self._agent_turn = agent_turn
        self._agent = agent

    def handle_events(self, event):
        env = self.env.unwrapped
        assert isinstance(env, AlkkagiEnv)

        if self._agent_turn is not None and self._agent is not None:
            if self._agent_turn == env.turn_:
                return self._agent.act(env.latest_obs_, env.latest_info_)

        if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1 and not self._is_dragging:
            mouse_pos = pygame.mouse.get_pos()
            if not self._is_dragging:
                self._selected_stone = env.find_stone(mouse_pos)
                if self._selected_stone is not None:
                    self._is_dragging = True
                    self._pos_start = self._selected_stone.position_
                    self._pos_end = mouse_pos
        elif event.type == pygame.MOUSEBUTTONUP and event.button == 1 and self._is_dragging:
            sx, sy = self._pos_start
            ex, ey = self._pos_end
            dx, dy = sx - ex, sy - ey
            angle = pygame.math.Vector2(1, 0).angle_to(pygame.math.Vector2(dx, dy))
            distance = math.hypot(dx, dy)
            power = min(distance * env.max_power / SlingShot.POWER_LINE_LENGTH, env.max_power)
            action = {
                'turn': env.turn_,
                'angle': angle,
                'power': power,
                'index': self._selected_stone.index_
            }
            self._is_dragging = False
            self._pos_start = None
            self._pos_end = None
            self._selected_stone = None
            return action
        elif event.type == pygame.MOUSEMOTION and self._is_dragging:
            self._pos_end = pygame.mouse.get_pos()

        env.set_slingshot_from_pos(self._pos_start, self._pos_end)

        return None


class RGBImgObsWrapper(gym.ObservationWrapper):
    def __init__(self, env: gym.Env):
        super().__init__(env)
        env = env.unwrapped
        assert isinstance(env, AlkkagiEnv)
        self.observation_space = gym.spaces.Dict({
            'image': gym.spaces.Box(
                low=0, high=255, shape=(env.width, env.height, 3), dtype=np.uint8
            ),
            'turn': gym.spaces.Discrete(2),
        })

    def observation(self, observation):
        env = self.env.unwrapped
        assert isinstance(env, AlkkagiEnv)
        img = env.get_frame()
        return {
            'image': img,
            'turn': env.turn_,
        }


class AlkkagiRemoteEnvWrapper(RemoteEnvWrapper):
    def serialize(self, observation: Any):
        return observation

    def verify_action(self, action) -> bool:
        env = self.env.unwrapped
        return isinstance(env, AlkkagiEnv) and action['turn'] == env.turn_



from .wrappers import AlkkagiManualPlayWrapper as ManualPlayWrapper
from .wrappers import AlkkagiRemoteEnvWrapper as RemoteEnvWrapper, RGBImgObsWrapper

__all__ = [
    'ManualPlayWrapper',
    'RemoteEnvWrapper'
]



"""
kymnasium - A reinforcement learning environment package

This package provides custom environments for reinforcement learning experiments.
"""

import os
import gymnasium as gym
from .agent import Agent
from .evaluate import evaluate, evaluate_remote, RemoteEnvWrapper, InvalidActionError, NotAllowedUserIdError
from . import alkkagi, avoid_blurp, grid_world


__all__ = [
    # Core classes and methods
    'Agent',
    'evaluate',
    'evaluate_remote',
    'RemoteEnvWrapper',
    'InvalidActionError',
    'NotAllowedUserIdError',
    # Modules
    'alkkagi',
    'avoid_blurp',
    'grid_world',
]

os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"

__version__ = "1.1.3"



# Al-Kka-Gi
# -------------------------------------------------------------------------------------------
gym.register(
    id='kymnasium/AlKkaGi-3x3-v0',
    entry_point='kymnasium.alkkagi.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        n_stones=3,
        n_obstacles=3
    )
)

gym.register(
    id='kymnasium/AlKkaGi-5x5-v0',
    entry_point='kymnasium.alkkagi.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        n_stones=5,
        n_obstacles=3
    )
)

gym.register(
    id='kymnasium/AlKkaGi-7x7-v0',
    entry_point='kymnasium.alkkagi.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        n_stones=7,
        n_obstacles=3
    )
)

gym.register(
    id='kymnasium/AlKkaGi-9x9-v0',
    entry_point='kymnasium.alkkagi.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        n_stones=9,
        n_obstacles=3
    )
)
# -------------------------------------------------------------------------------------------


# Avoid Blurp
# -------------------------------------------------------------------------------------------

gym.register(
    id='kymnasium/AvoidBlurp-Discrete-Vertical-Easy-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.5,
        max_spawns=30,
        prob_spawn_on_player=0.0,
        max_spawn_duration=105,
        continuous_action=False,
        mode='vertical',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Discrete-Vertical-Normal-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.3,
        max_spawns=30,
        prob_spawn_on_player=0.1,
        max_spawn_duration=105,
        continuous_action=False,
        mode='vertical',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Discrete-Vertical-Hard-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.1,
        max_spawns=30,
        prob_spawn_on_player=0.2,
        max_spawn_duration=105,
        continuous_action=False,
        mode='vertical',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Continuous-Vertical-Easy-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.5,
        max_spawns=30,
        prob_spawn_on_player=0.0,
        max_spawn_duration=105,
        continuous_action=True,
        mode='vertical',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Continuous-Vertical-Normal-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.3,
        max_spawns=30,
        prob_spawn_on_player=0.1,
        max_spawn_duration=105,
        continuous_action=True,
        mode='vertical',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Continuous-Vertical-Hard-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.1,
        max_spawns=30,
        prob_spawn_on_player=0.2,
        max_spawn_duration=105,
        continuous_action=True,
        mode='vertical',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Discrete-Ballistic-Easy-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.5,
        max_spawns=30,
        prob_spawn_on_player=0.0,
        max_spawn_duration=105,
        continuous_action=False,
        mode='ballistic',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Discrete-Ballistic-Normal-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.3,
        max_spawns=30,
        prob_spawn_on_player=0.1,
        max_spawn_duration=105,
        continuous_action=False,
        mode='ballistic',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Discrete-Ballistic-Hard-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.1,
        max_spawns=30,
        prob_spawn_on_player=0.2,
        max_spawn_duration=105,
        continuous_action=False,
        mode='ballistic',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Continuous-Ballistic-Easy-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.5,
        max_spawns=30,
        prob_spawn_on_player=0.0,
        max_spawn_duration=105,
        continuous_action=True,
        mode='ballistic',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Continuous-Ballistic-Normal-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.3,
        max_spawns=30,
        prob_spawn_on_player=0.1,
        max_spawn_duration=105,
        continuous_action=True,
        mode='ballistic',
        stage=1
    )
)

gym.register(
    id='kymnasium/AvoidBlurp-Continuous-Ballistic-Hard-Stage-1',
    entry_point='kymnasium.avoid_blurp.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.1,
        max_spawns=30,
        prob_spawn_on_player=0.2,
        max_spawn_duration=105,
        continuous_action=True,
        mode='ballistic',
        stage=1
    )
)
# -------------------------------------------------------------------------------------------


# Bullet Bill
# -------------------------------------------------------------------------------------------

gym.register(
    id='kymnasium/BulletBill-Discrete-Easy-Stage-1',
    entry_point='kymnasium.bullet_bill.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.5,
        max_spawns=30,
        max_spawn_duration=105,
        continuous_action=False,
    )
)

gym.register(
    id='kymnasium/BulletBill-Discrete-Normal-Stage-1',
    entry_point='kymnasium.bullet_bill.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.3,
        max_spawns=30,
        max_spawn_duration=105,
        continuous_action=False,
    )
)

gym.register(
    id='kymnasium/BulletBill-Discrete-Hard-Stage-1',
    entry_point='kymnasium.bullet_bill.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.1,
        max_spawns=30,
        max_spawn_duration=105,
        continuous_action=False,
    )
)

gym.register(
    id='kymnasium/BulletBill-Continuous-Easy-Stage-1',
    entry_point='kymnasium.bullet_bill.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.5,
        max_spawns=30,
        max_spawn_duration=105,
        continuous_action=True,
    )
)

gym.register(
    id='kymnasium/BulletBill-Continuous-Normal-Stage-1',
    entry_point='kymnasium.bullet_bill.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.3,
        max_spawns=30,
        max_spawn_duration=105,
        continuous_action=True,
    )
)

gym.register(
    id='kymnasium/BulletBill-Continuous-Hard-Stage-1',
    entry_point='kymnasium.bullet_bill.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        game_duration=120,
        init_spawn_interval=1.5,
        min_spawn_interval=0.1,
        max_spawns=30,
        max_spawn_duration=105,
        continuous_action=True,
    )
)

# Grid World
# -------------------------------------------------------------------------------------------
gym.register(
    id='kymnasium/GridWorld-Adventure-26x26',
    entry_point='kymnasium.grid_world.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        max_steps=500,
        blueprint=os.path.join(
            os.path.dirname(__file__),
            'grid_world',
            'assets',
            'adventure-26x26.csv'
        ),
    )
)

gym.register(
    id='kymnasium/GridWorld-Adventure-32x32',
    entry_point='kymnasium.grid_world.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        max_steps=500,
        blueprint=os.path.join(
            os.path.dirname(__file__),
            'grid_world',
            'assets',
            'adventure-32x32.csv'
        ),
    )
)

gym.register(
    id='kymnasium/GridWorld-Crossing-26x26',
    entry_point='kymnasium.grid_world.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        max_steps=500,
        blueprint=os.path.join(
            os.path.dirname(__file__),
            'grid_world',
            'assets',
            'crossing-26x26.csv'
        ),
    )
)



# Zelda Adventure
# -------------------------------------------------------------------------------------------
gym.register(
    id='kymnasium/ZeldaAdventure-Stage-1',
    entry_point='kymnasium.zelda_adventure.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        max_steps=500,
        stage=1
    )
)

gym.register(
    id='kymnasium/ZeldaAdventure-Stage-2',
    entry_point='kymnasium.zelda_adventure.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        max_steps=1000,
        stage=2
    )
)

gym.register(
    id='kymnasium/ZeldaAdventure-Stage-3',
    entry_point='kymnasium.zelda_adventure.registration:_create_env',
    disable_env_checker=True,
    kwargs=dict(
        max_steps=1000,
        stage=3
    )
)



class InvalidActionError(Exception):
    def __init__(self, action):
        self._action = action

    def __str__(self):
        return f"Invalid action: {self._action}"


class NotAllowedUserIdError(Exception):
    def __init__(self, user_id):
        self._user_id = user_id

    def __str__(self):
        return f"Not allowed user id: {self._user_id}"


class ObservationError(Exception):
    def __init__(self, message: str):
        self._message = message

    def __str__(self):
        return self._message

class EnvironmentClosedError(Exception):
    def __init__(self, message: str):
        self._message = message

    def __str__(self):
        return self._message